# Music Applet
# Copyright (C) 2008 Paul Kuliniewicz <paul@kuliniewicz.org>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02111-1301, USA.

from gettext import gettext as _
import gobject
import gtk
from gtk import gdk, glade

import os
import sys
import threading

import musicapplet.defs
import musicapplet.player


INFO = {
        "name": "Amarok",
        "version": musicapplet.defs.VERSION,
        "icon-name": "amarok",
        "author": "Paul Kuliniewicz <paul@kuliniewicz.org>",
        "copyright": "(C) 2007-2008 Paul Kuliniewicz",
        "website": "http://www.kuliniewicz.org/music-applet/",
}


class AmarokPlugin (musicapplet.player.PluginBase):
    """
    Music Applet plugin to interface with Amarok.
    """


    ##################################################################
    #
    # Creation and connection management
    #
    ##################################################################


    def __init__ (self, conf):
        musicapplet.player.PluginBase.__init__ (self, conf, INFO["name"], INFO["icon-name"])

        import kdecore
        import dcopext

        self.min_rating = 0.0
        self.max_rating = 5.0

        self._app = None
        self._app_lock = threading.Lock ()

        self._should_start = False
        self._should_start_lock = threading.Lock ()

        self.__amarok = None
        self.__poll_source = None
        self.__url = None

    def start (self):
        # KApplication is needed to reliably start the DCOP server and to create the
        # wrapper object for Amarok's DCOP interface.  It's created in a separate
        # thread because it can take a long time to create.  The locking assures only
        # one KApplication ever gets created, even if start() is called repeatedly
        # while the other thread is working.

        if self._app_lock.acquire (False):
            if self._app is None:
                self._should_start = True       # lock unnecessary; no other thread running
                thread = AmarokInitThread (self)
                thread.start()
            else:
                self._start_safely ()
                self._app_lock.release ()

    def _start_safely (self):
        if self.__amarok is None:
            import dcopext
            self.__amarok = dcopext.DCOPApp ("amarok", self._app.dcopClient ())

        if self.__poll_source is None and self.__poll_connect ():
            self.__poll_source = gobject.timeout_add (1000, self.__poll_connect)

    def stop (self):
        if self.__poll_source is not None:
            gobject.source_remove (self.__poll_source)
            self.__poll_source = None

        self._should_start_lock.acquire (True)
        self._should_start = False
        self._should_start_lock.release ()

        # Leave _app alone; KDE doesn't like it when you try creating more than one.

        self._set_no_song ()
        self.__amarok = None
        self.connected = False

    def __poll_connect (self):
        ok, dummy = self.__amarok.player.title ()       # something to actually test for a connection
        if ok:
            self.connected = True
            self.__poll_status ()
            self.__poll_source = gobject.timeout_add (1000, self.__poll_status)
            return False
        else:
            return True

    def __safe_call (self, method_name):
        """
        Call a DCOP method on the player, throwing an exception if it fails.
        """

        ok, result = self.__amarok.player.method (method_name) ()
        if ok:
            return result
        else:
            raise RuntimeError


    ##################################################################
    #
    # Playback controls
    #
    ##################################################################


    def toggle_playback (self):
        self.__amarok.player.playPause ()

    def previous (self):
        self.__amarok.player.prev ()

    def next (self):
        self.__amarok.player.next ()

    def rate_song (self, rating):
        self.__amarok.player.setRating (int (rating * 2))


    ##################################################################
    #
    # Status
    #
    ##################################################################

    def __poll_status (self):
        try:
            # TODO: when dcopext supports async calls, use them instead
            url = self.__safe_call ("encodedURL") or None
            if url != self.__url:
                self.__url = url
                if self.__url is not None:
                    self.title = self.__safe_call ("title") or None
                    self.album = self.__safe_call ("album") or None
                    self.artist = self.__safe_call ("artist") or None
                    self.duration = self.__safe_call ("trackTotalTime")
                    self.rating = self.__safe_call ("rating") / 2.0
                    art_file = self.__safe_call ("coverImage")
                    if not art_file.endswith ("@nocover.png"):
                        self.art = gdk.pixbuf_new_from_file (art_file)
                    else:
                        self.art = None
                else:
                    self._set_no_song ()

            self.playing = self.__safe_call ("isPlaying")
            self.elapsed = self.__safe_call ("trackCurrentTime")

            return True

        except RuntimeError:
            self._set_no_song ()
            self.connected = False
            self.__poll_source = gobject.timeout_add (1000, self.__poll_connect)
            return False


    ##################################################################
    #
    # Configuration
    #
    ##################################################################

    def create_config_dialog (self):
        xml = glade.XML (os.path.join (musicapplet.defs.PKG_DATA_DIR, "amarok.glade"))
        dialog = xml.get_widget ("amarok-dialog")

        full_key = self._conf.resolve_plugin_key (self, "command")
        self._conf.bind_string (full_key, xml.get_widget ("command"), "text")

        xml.get_widget ("browse").connect ("clicked", lambda button: self._browse_for_command (dialog, "Amarok"))
        dialog.set_title (_("%s Plugin") % "Amarok")
        dialog.set_default_response (gtk.RESPONSE_CLOSE)
        dialog.connect ("response", lambda dialog, response: dialog.hide ())
        return dialog


class AmarokInitThread (threading.Thread):
    """
    Separate thread that initializes the plugin's KApplication object.  This
    is done in a thread because the call can take several seconds to complete
    if the KDE services are not already running.

    When thread execution begins, it inherits the _app_lock of the plugin.
    """

    def __init__ (self, plugin):
        threading.Thread.__init__ (self, name="Amarok Init Thread")
        self.__plugin = plugin

    def run (self):
        # The KApplication constructor likes to call exit() if it finds
        # arguments it doesn't understand in argv -- such as the arguments
        # Bonobo uses to launch the applet -- so hide everything but the
        # program name from it.

        import kdecore
        fake_argv = sys.argv[0:1]
        self.__plugin._app = kdecore.KApplication (fake_argv, "music-applet")

        # If a non-Amarok player was already running at applet startup, it's
        # probable that start() and stop() were both called before getting
        # to this point in the thread.

        self.__plugin._should_start_lock.acquire (True)
        if self.__plugin._should_start:
            self.__plugin._start_safely ()
            self.__plugin._should_start = False
        self.__plugin._should_start_lock.release ()

        self.__plugin._app_lock.release ()


def create_instance (conf):
    return AmarokPlugin (conf)



# Music Applet
# Copyright (C) 2007 Paul Kuliniewicz <paul@kuliniewicz.org>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02111-1301, USA.

"""
Classes that provide access to the applet's configuration.

An instance of the Conf class is provided to each plugin, giving it
access to the applet's preferences as well as the preferences specific
to that plugin.  All preferences are stored in GConf internally.
"""

import gconf
import gtk
from gtk import glade

import os

import musicapplet.defs
import musicapplet.util


class Conf:
    """
    Helper object for managing the applet's preferences.

    The "client" object is a convenience instance of a GConf client, which
    plugin writers will use directly most of the time.  The Conf object
    itself is primarily useful in mapping keys into where they're actually
    stored in GConf, since each instance of the applet has its preferences
    stored in a different location.
    
    There are also methods that bind a GConf entry to some other object, so
    that changes in one automatically cause a change in the other.
    """


    # For convenience
    client = gconf.client_get_default ()


    def __init__ (self, applet):
        self.__applet = applet
        print "DEBUG: GConf root is %s" % applet.get_preferences_key ()


    def bind_boolean (self, full_key, object, property):
        """
        Bind a GConf boolean value to a GObject's property.
        """

        def gconf_cb (client, id, entry, unused):
            if entry.value is not None and entry.value.get_bool () != object.get_property (property):
                object.set_property (property, entry.value.get_bool ())

        def object_cb (object, pspec):
            if object.get_property (property) != self.client.get_bool (full_key):
                self.client.set_bool (full_key, object.get_property (property))

        def destroy_cb (object, id):
            self.client.notify_remove (id)

        object.set_property (property, self.client.get_bool (full_key))
        id = self.client.notify_add (full_key, gconf_cb)
        object.connect ("notify::%s" % property, object_cb)
        object.connect ("destroy", destroy_cb, id)


    def bind_int (self, full_key, object, property):
        """
        Bind a GConf int value to a GObject's property.
        """

        def gconf_cb (client, id, entry, unused):
            if entry.value is not None and entry.value.get_int () != object.get_property (property):
                object.set_property (property, entry.value.get_int ())

        def object_cb (object, pspec):
            if object.get_property (property) != self.client.get_int (full_key):
                self.client.set_int (full_key, int (object.get_property (property)))

        def destroy_cb (object, id):
            self.client.notify_remove (id)

        object.set_property (property, self.client.get_int (full_key))
        id = self.client.notify_add (full_key, gconf_cb)
        object.connect ("notify::%s" % property, object_cb)
        object.connect ("destroy", destroy_cb, id)


    def bind_string (self, full_key, object, property):
        """
        Bind a GConf string value to a GObject's property.
        """

        def gconf_cb (client, id, entry, unused):
            if entry.value is not None and entry.value.get_string () != object.get_property (property):
                object.set_property (property, entry.value.get_string ())

        def object_cb (object, pspec):
            if object.get_property (property) != self.client.get_string (full_key):
                self.client.set_string (full_key, object.get_property (property))

        def destroy_cb (object, id):
            self.client.notify_remove (id)

        object.set_property (property, self.client.get_string (full_key))
        id = self.client.notify_add (full_key, gconf_cb)
        object.connect ("notify::%s" % property, object_cb)
        object.connect ("destroy", destroy_cb, id)


    def bind_string_boolean (self, full_key, value, object, property):
        """
        Bind a GConf string value to a GObject's boolean property, by
        setting the property to true iff the string matches a particular
        value, and vice versa.
        """

        def gconf_cb (client, id, entry, unused):
            if entry.value is not None:
                bval = (entry.value.get_string () == value)
                if bval != object.get_property (property):
                    object.set_property (property, bval)

        def object_cb (object, pspec):
            if object.get_property (property) and self.client.get_string (full_key) != value:
                self.client.set_string (full_key, value)

        def destroy_cb (object, id):
            self.client.notify_remove (id)

        object.set_property (property, self.client.get_string (full_key) == value)
        id = self.client.notify_add (full_key, gconf_cb)
        object.connect ("notify::%s" % property, object_cb)
        object.connect ("destroy", destroy_cb, id)


    def bind_combo_string (self, full_key, combo, column):
        """
        Bind a GConf string value to a ComboBox column.
        """

        def gconf_cb (client, id, entry, unused):
            if entry.value is not None:
                iter = musicapplet.util.search_model (combo.get_model (), column, entry.value.get_string ())
                if iter is not None and iter != combo.get_active_iter ():
                    combo.set_active_iter (iter)

        def combo_cb (combo):
            model = combo.get_model ()
            value = model.get_value (combo.get_active_iter (), column)
            if value != self.client.get_string (full_key):
                self.client.set_string (full_key, value)

        def destroy_cb (combo, id):
            self.client.notify_remove (id)

        iter = musicapplet.util.search_model (combo.get_model (), column, self.client.get_string (full_key))
        if iter is not None:
            combo.set_active_iter (iter)
        id = self.client.notify_add (full_key, gconf_cb)
        combo.connect ("changed", combo_cb)
        combo.connect ("destroy", destroy_cb, id)


    def resolve_key (self, key):
        """
        Resolve the full name of a key for general applet preferences.
        """

        return "%s/%s" % (self.__applet.get_preferences_key (), key)


    def resolve_plugin_key (self, plugin, key):
        """
        Resolve the full name of a key for a particular plugin.

        This function will try the preferred key style "Player-Name/key",
        falling back on the legacy, deprecated style "player-name_key".
        Ultimately, the schema determines which will be used.
        """

        escaped = plugin.internal_name.replace (" ", "-")
        legacy_key = self.resolve_key ("%s_%s" % (escaped.lower (), key))
        if self.client.get (legacy_key) is not None:
            return legacy_key
        else:
            return self.resolve_key ("%s/%s" % (escaped, key))


def create_preferences_dialog (conf):
    """
    Create a dialog for configuring general applet preferences.
    """

    xml = glade.XML (os.path.join (musicapplet.defs.PKG_DATA_DIR, "preferences.glade"))

    full_key = conf.resolve_key ("show_song_information")
    conf.bind_boolean (full_key, xml.get_widget ("show-song-information"), "active")

    full_key = conf.resolve_key ("show_rating")
    conf.bind_boolean (full_key, xml.get_widget ("show-rating"), "active")

    full_key = conf.resolve_key ("show_time")
    conf.bind_boolean (full_key, xml.get_widget ("show-time"), "active")

    full_key = conf.resolve_key ("show_controls")
    conf.bind_boolean (full_key, xml.get_widget ("show-controls"), "active")

    full_key = conf.resolve_key ("show_notification")
    conf.bind_boolean (full_key, xml.get_widget ("show-notification"), "active")
    try:
        import pynotify
    except ImportError:
        xml.get_widget ("show-notification").set_sensitive (False)

    xml.get_widget ("preferences-dialog").set_default_response (gtk.RESPONSE_CLOSE)
    xml.get_widget ("preferences-dialog").connect ("response", lambda dialog, response: dialog.hide ())
    return xml.get_widget ("preferences-dialog")



# Music Applet
# Copyright (C) 2007 Paul Kuliniewicz <paul@kuliniewicz.org>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02111-1301, USA.

import gobject
from gtk import gdk

import musicapplet.defs
import musicapplet.player


INFO = {
        "name": "Rhythmbox",
        "version": musicapplet.defs.VERSION,
        "icon-name": "rhythmbox",
        "author": "Paul Kuliniewicz <paul@kuliniewicz.org>",
        "copyright": "(C) 2007 Paul Kuliniewicz",
        "website": "http://www.kuliniewicz.org/music-applet/",
}


class RhythmboxPlugin (musicapplet.player.DBusPlugin):
    """
    Music Applet plugin to interface with Rhythmbox.
    """


    ##################################################################
    #
    # Creation and overall management
    #
    ##################################################################


    def __init__ (self, conf):
        musicapplet.player.DBusPlugin.__init__ (self, conf, "Rhythmbox", "rhythmbox",
                                                "org.gnome.Rhythmbox")

        self.min_rating = 0.0
        self.max_rating = 5.0

        self.__handlers = []
        self.__uri = None

        self.__player = None
        self.__shell = None


    ##################################################################
    #
    # Playback control
    #
    ##################################################################


    def toggle_playback (self):
        self.__player.playPause (False,
                                 reply_handler=lambda: None,
                                 error_handler=self._report_dbus_error)


    def previous (self):
        self.__player.previous (reply_handler=lambda: None,
                                error_handler=self._report_dbus_error)


    def next (self):
        self.__player.next (reply_handler=lambda: None,
                            error_handler=self._report_dbus_error)


    ##################################################################
    #
    # Elapsed time
    #
    ##################################################################


    def __set_elapsed (self, elapsed):
        if elapsed > self.MAX_TIME:
            self.elapsed = -1
        else:
            self.elapsed = elapsed

        if not self.playing and self.elapsed < 0:
            self._set_no_song ()


    ##################################################################
    #
    # Title
    #
    ##################################################################


    def __set_title (self, title):
        self.title = title


    ##################################################################
    #
    # Artist
    #
    ##################################################################


    def __set_artist (self, artist):
        self.artist = artist


    ##################################################################
    #
    # Album
    #
    ##################################################################


    def __set_album (self, album):
        self.album = album


    ##################################################################
    #
    # Duration
    #
    ##################################################################


    def __set_duration (self, duration):
        self.duration = duration


    ##################################################################
    #
    # Playing
    #
    ##################################################################


    def __set_playing (self, playing):
        self.playing = playing

        if not self.playing and self.elapsed < 0:
            self._set_no_song ()


    ##################################################################
    #
    # URI
    #
    ##################################################################


    def __set_uri (self, uri):
        if uri is not None and uri != "":
            self.__uri = uri
            self.__shell.getSongProperties (uri,
                                            reply_handler=self.__set_song_properties,
                                            error_handler=self._report_dbus_error)
        else:
            self._set_no_song ()


    def __set_song_properties (self, properties):
        self.title = properties.get ("title")
        self.artist = properties.get ("artist")
        self.album = properties.get ("album")
        self.duration = properties.get ("duration", -1)
        self.rating = properties.get ("rating", -1.0)
        if properties.has_key ("rb:coverArt-uri"):
            self.art = gdk.pixbuf_new_from_file (properties.get ("rb:coverArt-uri"))
        else:
            self.art = None


    def __property_changed (self, uri, property, old_value, new_value):
        if property == "title":
            self.title = new_value
        elif property == "artist":
            self.artist = new_value
        elif property == "album":
            self.album = new_value
        elif property == "duration":
            self.duration = new_value
        elif property == "rating":
            self.rating = new_value
        elif property == "rb:coverArt-uri":
            self.art = gdk.pixbuf_new_from_file (new_value)


    ##################################################################
    #
    # Rating
    #
    ##################################################################


    def __set_rating (self, rating):
        self.rating = rating


    def rate_song (self, rating):
        self.__shell.setSongProperty (self.__uri, "rating", float (rating),
                                      reply_handler=lambda: self.set_property ("rating", rating),
                                      error_handler=self._report_dbus_error)


    ##################################################################
    #
    # Connection management
    #
    ##################################################################


    def _connect (self):
        if self.__player is not None:
            return

        import dbus
        bus = dbus.SessionBus ()

        proxy = bus.get_object ("org.gnome.Rhythmbox", "/org/gnome/Rhythmbox/Player")
        self.__player = dbus.Interface (proxy, "org.gnome.Rhythmbox.Player")

        proxy = bus.get_object ("org.gnome.Rhythmbox", "/org/gnome/Rhythmbox/Shell")
        self.__shell = dbus.Interface (proxy, "org.gnome.Rhythmbox.Shell")

        self.__handlers = [
            self.__player.connect_to_signal ("playingChanged", self.__set_playing),
            self.__player.connect_to_signal ("elapsedChanged", self.__set_elapsed),
            self.__player.connect_to_signal ("playingUriChanged", self.__set_uri),
            self.__player.connect_to_signal ("playingSongPropertyChanged", self.__property_changed),
        ]

        self.__player.getPlaying (reply_handler=self.__set_playing,
                                  error_handler=self._report_dbus_error)

        self.__player.getElapsed (reply_handler=self.__set_elapsed,
                                  error_handler=self._report_dbus_error)

        self.__player.getPlayingUri (reply_handler=self.__set_uri,
                                     error_handler=self._report_dbus_error)

        self.connected = True


    def _disconnect (self):
        for handler in self.__handlers:
            handler.remove ()
        self.__handlers = []

        self.__player = None
        self.__shell = None

        self._set_no_song ()
        self.connected = False


def create_instance (conf):
    return RhythmboxPlugin (conf)



# Music Applet
# Copyright (C) 2007 Paul Kuliniewicz <paul@kuliniewicz.org>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02111-1301, USA.

import musicapplet.player
import musicapplet.widgets

import array

import gobject
from gtk import gdk


INFO = {
        "name": "Muine",
        "version": musicapplet.defs.VERSION,
        "icon-name": "muine",
        "author": "Paul Kuliniewicz <paul@kuliniewicz.org>",
        "copyright": "(C) 2007 Paul Kuliniewicz",
        "website": "http://www.kuliniewicz.org/music-applet/",
}


class MuinePlugin (musicapplet.player.DBusPlugin):
    """
    Music Applet plugin to interface with Muine.
    """


    ##################################################################
    #
    # Creation and overall management
    #
    ##################################################################


    def __init__ (self, conf):
        musicapplet.player.DBusPlugin.__init__ (self, conf, "Muine", "muine",
                                                "org.gnome.Muine")

        self.__player = None
        self.__handlers = []
        self.__poll_source = None


    ##################################################################
    #
    # Playback control
    #
    ##################################################################


    def toggle_playback (self):
        self.__player.SetPlaying (not self.playing,
                                  reply_handler=lambda: None,
                                  error_handler=self._report_dbus_error)


    def previous (self):
        self.__player.Previous (reply_handler=lambda: None,
                                error_handler=self._report_dbus_error)


    def next (self):
        self.__player.Next (reply_handler=lambda: None,
                            error_handler=self._report_dbus_error)


    ##################################################################
    #
    # Playing
    #
    ##################################################################


    def __set_playing (self, playing):
        self.playing = playing


    ##################################################################
    #
    # Song Data
    #
    ##################################################################


    def __set_song (self, info):
        if len (info) > 0:
            artists = []
            self.album = None

            for line in info.splitlines ():
                [name, value] = line.split (": ", 1)
                if name == "title":
                    self.title = value
                elif name == "artist" or name == "performer":
                    artists.append (value)
                elif name == "album":
                    self.album = value
                elif name == "duration":
                    self.duration = int (value)

            if len (artists) > 0:
                self.artist = ", ".join (artists)
            else:
                self.artist = None

            # Muine crashes if you call GetAlbumCover and there isn't any
            # art available, but the result of saving it to /dev/null will
            # tell us whether it exists.
            #
            # See http://bugzilla.gnome.org/show_bug.cgi?id=396624

            self.__player.WriteAlbumCoverToFile ("/dev/null",
                                                 reply_handler=self.__does_art_exist,
                                                 error_handler=self._report_dbus_error)

        else:
            self._set_no_song ()


    ##################################################################
    #
    # Art
    #
    ##################################################################


    def __does_art_exist (self, exists):
        if exists:
            self.__player.GetAlbumCover (reply_handler=self.__set_art,
                                         error_handler=self._report_dbus_error)
        else:
            self.art = None


    def __set_art (self, bytes):
        raw = array.array ('B', [int(b) for b in bytes])
        if len (raw) > 0:
            self.art = musicapplet.widgets.ma_deserialize_pixdata (raw.tostring ())
        else:
            self.art = None


    ##################################################################
    #
    # Elapsed
    #
    ##################################################################


    def __set_elapsed (self, elapsed):
        if self.title is not None:
            self.elapsed = elapsed


    ##################################################################
    #
    # Connection management
    #
    ##################################################################


    def _connect (self):
        if self.__player is not None:
            return

        import dbus
        bus = dbus.SessionBus ()

        proxy = bus.get_object ("org.gnome.Muine", "/org/gnome/Muine/Player")
        self.__player = dbus.Interface (proxy, "org.gnome.Muine.Player")

        self.__handlers = [
            self.__player.connect_to_signal ("StateChanged", self.__set_playing),
            self.__player.connect_to_signal ("SongChanged", self.__set_song),
        ]

        self.__player.GetPlaying (reply_handler=self.__set_playing,
                                  error_handler=self._report_dbus_error)

        self.__player.GetCurrentSong (reply_handler=self.__set_song,
                                      error_handler=self._report_dbus_error)

        self.__poll ()
        self.connected = True
        self.__poll_source = gobject.timeout_add (1000, self.__poll)


    def _disconnect (self):
        if self.__poll_source is not None:
            gobject.source_remove (self.__poll_source)
            self.__poll_source = None

        for handler in self.__handlers:
            handler.remove ()
        self.__handlers = []

        self.__player = None

        self._set_no_song ()
        self.connected = False


    def __poll (self):
        self.__player.GetPosition (reply_handler=self.__set_elapsed,
                                   error_handler=self._report_dbus_error)

        return True


def create_instance (conf):
    return MuinePlugin (conf)



# Music Applet
# Copyright (C) 2007 Paul Kuliniewicz <paul@kuliniewicz.org>
#           (C) 2007 Mickael Royer <mickael.royer@gmail.com>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02111-1301, USA.

import musicapplet.defs
import musicapplet.player

import gobject


INFO = {
        "name": "Quod Libet",
        "version": musicapplet.defs.VERSION,
        "icon-name": "quodlibet",
        "author": "Mickael Royer <mickael.royer@gmail.com>",
        "copyright": "(C) 2007 Mickael Royer",
        "website": "http://www.kuliniewicz.org/music-applet/",
}


class QuodLibetPlugin (musicapplet.player.DBusPlugin):

    def __init__ (self, conf):
        musicapplet.player.DBusPlugin.__init__ (self, conf,
                                                INFO["name"], INFO["icon-name"],
                                                "net.sacredchao.QuodLibet")

        self.min_rating = 0.0
        self.max_rating = 4.0

        self.__player = None
        self.__handlers = []
        self.__poll_source = None


    def _connect (self):
        if self.__player is not None:
            return

        import dbus
        bus = dbus.SessionBus ()

        proxy = bus.get_object ("net.sacredchao.QuodLibet",
                                "/net/sacredchao/QuodLibet")
        self.__player = dbus.Interface (proxy, "net.sacredchao.QuodLibet")

        self.__handlers = [
            self.__player.connect_to_signal ("SongStarted", self.__song_started),
            self.__player.connect_to_signal ("SongEnded", self.__song_ended),
            self.__player.connect_to_signal ("Paused", self.__set_stop),
            self.__player.connect_to_signal ("Unpaused", self.__set_start),
        ]

        self.__player.CurrentSong (reply_handler=self.__set_song,
                                   error_handler=self._report_dbus_error)

        self.__poll ()
        self.connected = True
        self.__poll_source = gobject.timeout_add (1000, self.__poll)


    def _disconnect (self):
        if self.__poll_source is not None:
            gobject.source_remove (self.__poll_source)
            self.__poll_source = None

        for handler in self.__handlers:
            handler.remove ()
        self.__handlers = []

        self.__player = None
        self._set_no_song ()
        self.connected = False


    def __poll (self):
        self.__player.IsPlaying (reply_handler=self.__set_playing,
                                 error_handler=self._report_dbus_error)
        self.__player.GetPosition (reply_handler=self.__set_elapsed,
                                   error_handler=self._report_dbus_error)

        return True


    def toggle_playback (self):
        self.__player.PlayPause (reply_handler=lambda paused: None,
                                 error_handler=self._report_dbus_error)


    def previous (self):
        self.__player.Previous (reply_handler=lambda: None,
                                error_handler=self._report_dbus_error)


    def next (self):
        self.__player.Next (reply_handler=lambda: None,
                            error_handler=self._report_dbus_error)


    def __set_start (self):
        self.playing = True


    def __set_stop (self):
        self.playing = False


    def __set_playing (self, playing):
        self.playing = playing


    def __song_started (self, song):
        self.__set_song (song)


    def __song_ended (self, song, skipped):
        self._set_no_song ()


    def __set_song (self, song):
        if len (song) == 0:
            self._set_no_song ()
        else:
            if "title" in song.keys ():
                self.title = song["title"]
            if "artist" in song.keys ():
                self.artist = song["artist"]
            if "album" in song.keys ():
                self.album = song["album"]
            if "~#length" in song.keys ():
                self.duration = int (song["~#length"])
            if "~#rating" in song.keys ():
                self.rating = int (float (song["~#rating"]) * 4)


    def __set_elapsed (self, elapsed):
        if self.title is not None:
            self.elapsed = int (elapsed / 1000)
        else:
            self.elapsed = -1


def create_instance (conf):
    return QuodLibetPlugin (conf)



# Music Applet
# Copyright (C) 2007 Paul Kuliniewicz <paul@kuliniewicz.org>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02111-1301, USA.

import musicapplet.defs
import musicapplet.player

import gobject
from gtk import gdk

import re


INFO = {
        "name": "Exaile",
        "version": musicapplet.defs.VERSION,
        "icon-name": "exaile",
        "author": "Paul Kuliniewicz <paul@kuliniewicz.org>",
        "copyright": "(C) 2007 Paul Kuliniewicz",
        "website": "http://www.kuliniewicz.org/music-applet/",

}


class ExailePlugin (musicapplet.player.DBusPlugin):

    WEDGED = 3


    def __init__ (self, conf):
        musicapplet.player.DBusPlugin.__init__ (self, conf,
                                                INFO["name"], INFO["icon-name"],
                                                "org.exaile.DBusInterface")

        self.min_rating = 1.0
        self.max_rating = 8.0

        self.__player = None
        self.__art_path = None
        self.__poll_source = None

        self.__last_status = None
        self.__last_metadata = None
        self.__wedge_count = 0

        # Since matches are maximal, part 2 is the song metadata blurb
        self.__status_regex = re.compile ("^status: ([^ ]*) (.*) length: ([\d:]+) position: %\d+ \[([\d:]+)\]$")


    def _connect (self):
        if self.__player is not None:
            return

        import dbus
        bus = dbus.SessionBus ()

        proxy = bus.get_object ("org.exaile.DBusInterface", "/DBusInterfaceObject")
        self.__player = dbus.Interface (proxy, "org.exaile.DBusInterface")

        self.__poll ()
        self.connected = True
        self.__poll_source = gobject.timeout_add (1000, self.__poll)


    def _disconnect (self):
        self.__player = None

        if self.__poll_source is not None:
            gobject.source_remove (self.__poll_source)
            self.__poll_source = None

        self._set_no_song ()
        self.connected = False


    def toggle_playback (self):
        self.__player.play_pause (reply_handler=lambda: None,
                                  error_handler=self._report_dbus_error)


    def previous (self):
        self.__player.prev_track (reply_handler=lambda: None,
                                  error_handler=self._report_dbus_error)


    def next (self):
        self.__player.next_track (reply_handler=lambda: None,
                                  error_handler=self._report_dbus_error)


    def rate_song (self, rating):
        self.__player.set_rating (rating,
                                  reply_handler=lambda: self.set_property ("rating", rating),
                                  error_handler=self._report_dbus_error)


    def __poll (self):
        self.__player.query (reply_handler=self.__set_status,
                             error_handler=self._report_dbus_error)
        return True


    def __set_status (self, query_result):
        # Exaile 0.2.8 (and maybe others) gets "stuck" at the end of a playlist, so
        # try to detect this condition and react accordingly.

        if self.__last_status == query_result and query_result.startswith ("status: playing"):
            if self.__wedge_count < self.WEDGED:
                self.__wedge_count = self.__wedge_count + 1
        else:
            self.__last_status = query_result
            self.__wedge_count = 0

        match = self.__status_regex.match (query_result)
        if match is None or self.__wedge_count >= self.WEDGED:
            self._set_no_song ()
        else:
            self.playing = (match.group (1) == "playing")
            self.elapsed = self.__convert_time (match.group (4))
            if self.__last_metadata != match.group (2):
                self.__last_metadata = match.group (2)
                self.duration = self.__convert_time (match.group (3))

                # Only update these when they've changed.  Parsing them out of
                # the query result is unreliable in case strings like "album:"
                # actually appear in the metadata.

                self.__player.get_title (reply_handler=self.__set_title,
                                         error_handler=self._report_dbus_error)

                self.__player.get_artist (reply_handler=self.__set_artist,
                                          error_handler=self._report_dbus_error)

                self.__player.get_album (reply_handler=self.__set_album,
                                         error_handler=self._report_dbus_error)

                self.__player.get_cover_path (reply_handler=self.__set_art,
                                              error_handler=self._report_dbus_error)

                self.__player.get_rating (reply_handler=self.__set_rating,
                                          error_handler=self._report_dbus_error)


    def __set_title (self, title):
        if title != "":
            self.title = title
        else:
            self.title = None


    def __set_artist (self, artist):
        if artist != "":
            self.artist = artist
        else:
            self.artist = None


    def __set_album (self, album):
        if album != "":
            self.album = album
        else:
            self.album = None


    def __set_art (self, path):
        if path != "images/nocover.png" and path != self.__art_path:
            self.__art_path = path
            self.art = gdk.pixbuf_new_from_file (path)
        elif path == "images/nocover.png" and self.__art_path is not None:
            self.__art_path = None
            self.art = None


    def __set_rating (self, rating):
        self.rating = rating


    def __convert_time (self, time_str):
        # Convert an H:MM:SS or MM:SS string into a number of seconds.
        result = 0
        for part in time_str.split (":"):
            result = result * 60 + int (part)
        return result


def create_instance (conf):
    return ExailePlugin (conf)



# Music Applet
# Copyright (C) 2007 Paul Kuliniewicz <paul@kuliniewicz.org>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02111-1301, USA.

"""
A collection of utility methods that don't quite fit anywhere else.
"""


def search_model (model, column, value):
    """
    Search a TreeModel for the first row with a particular value in one
    column.
    """

    iter = model.get_iter_first ()
    while iter is not None:
        if model.get_value (iter, column) == value:
            return iter
        iter = model.iter_next (iter)
    return None


def make_property (name):
    """
    Create a Python property that wraps a GObject property.
    """

    return property (lambda self: self.get_property (name),
                     lambda self, value: self.set_property (name, value))



# Music Applet
# Copyright (C) 2007 Paul Kuliniewicz <paul@kuliniewicz.org>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02111-1301, USA.

"""
The graphical user interface of the applet itself.

Plugins shouldn't need to worry about anything here, as this is what
uses the plugins, not the other way around.
"""

from gettext import gettext as _
import gnomeapplet
import gobject
import gtk
from gtk import gdk
import pango

import os
from xml.sax import saxutils

import musicapplet.conf
import musicapplet.defs
import musicapplet.manager
import musicapplet.widgets

STOCK_ICON = "music-applet"
STOCK_SET_STAR = "music-applet-set-star"
STOCK_UNSET_STAR = "music-applet-unset-star"

USE_GTK_TOOLTIP = (gtk.check_version (2,12,0) == None)


class MusicApplet:
    """
    The actual applet part of Music Applet.
    """

    ##################################################################
    #
    # Construction
    #
    ##################################################################

    def __init__ (self, applet):

        # General

        gtk.window_set_default_icon_from_file (os.path.join (musicapplet.defs.PKG_DATA_DIR, "%s.png" % STOCK_ICON))

        self.__applet = applet
        self.__all = None

        self.__prefs_dialog = None
        self.__plugins_dialog = None
        self.__about_dialog = None
        self.__plugin_handlers = []

        applet.set_applet_flags (gnomeapplet.EXPAND_MINOR)

        applet.add_preferences ("/schemas/apps/music-applet/prefs")
        self.__conf = musicapplet.conf.Conf (applet)

        self.__manager = musicapplet.manager.Manager (self.__conf, [
            os.path.join (musicapplet.defs.PYTHON_DIR, "musicapplet/plugins"),
            os.path.expanduser ("~/.gnome2/music-applet/plugins"),
        ])

        try:
            self.__notify = NotifyPopup (applet, self.__conf)
        except ImportError:
            self.__notify = None

	# Tooltips

        self.__song_tip_guts = SongTooltipGuts ()
        if not USE_GTK_TOOLTIP:
            self.__tooltips = gtk.Tooltips ()

        if USE_GTK_TOOLTIP:
            self.__applet.props.has_tooltip = True
            self.__applet.connect ("query-tooltip", self.__query_tooltip_cb)
        else:
            song_tip = SongTooltip (self.__song_tip_guts)
            song_tip.set_tip (self.__applet, "")

        # Song scroller

        self.__scroller = musicapplet.widgets.Scroller ()
        self.__make_transparent (self.__scroller)
        full_key = self.__conf.resolve_key ("show_song_information")
        self.__conf.client.notify_add (full_key,
                lambda client, id, entry, unused:
                    entry.value is not None and self.__update_scroller_visibility (entry.value.get_bool ()))

        # Rating

        self.__rating = Rating ()
	self.__set_widget_tooltip (self.__rating, _("Rate current song"))
        self.__rating.connect ("notify::rating", self.__rate_song_cb)
        self.__make_transparent (self.__rating)
        full_key = self.__conf.resolve_key ("show_rating")
        self.__conf.client.notify_add (full_key,
                lambda client, id, entry, unused:
                    entry.value is not None and self.__update_rating_visibility (entry.value.get_bool ()))

        # Time display

        self.__time = gtk.Label ("--:--")
        self.__make_transparent (self.__time)
        full_key = self.__conf.resolve_key ("show_time")
        self.__conf.client.notify_add (full_key,
                lambda client, id, entry, unused:
                    entry.value is not None and self.__update_time_visibility (entry.value.get_bool ()))

        # Buttons

        self.__previous = musicapplet.widgets.FancyButton ()
        self.__previous.set_stock_id (gtk.STOCK_MEDIA_PREVIOUS)
	self.__set_widget_tooltip (self.__previous, _("Previous song"))
        self.__previous.connect ("clicked", self.__previous_clicked_cb)

        self.__play = musicapplet.widgets.FancyButton ()
        self.__play.set_stock_id (gtk.STOCK_MEDIA_PLAY)
	self.__set_widget_tooltip (self.__play, _("Start playing"))
        self.__play.connect ("clicked", self.__play_clicked_cb)

        self.__next = musicapplet.widgets.FancyButton ()
        self.__next.set_stock_id (gtk.STOCK_MEDIA_NEXT)
	self.__set_widget_tooltip (self.__next, _("Next song"))
        self.__next.connect ("clicked", self.__next_clicked_cb)

        full_key = self.__conf.resolve_key ("show_controls")
        self.__conf.client.notify_add (full_key,
                lambda client, id, entry, unused:
                    entry.value is not None and self.__update_controls_visibility (entry.value.get_bool ()))

        self.__launch = musicapplet.widgets.FancyButton ()
        self.__set_preferred (self.__manager.preferred)
        self.__launch.connect ("clicked", self.__launch_clicked_cb)

        # Context menu

        self.__applet.setup_menu_from_file (
            musicapplet.defs.LIB_DIR, "GNOME_Music_Applet.xml", None, [
                ("Preferences", self.__preferences_cb),
                ("Plugins", self.__plugins_cb),
                ("About", self.__about_cb),
            ])


        # Putting it all together

        self.__change_orient_cb (applet, applet.get_orient ())

        applet.connect ("change-background", self.__change_background_cb)
        applet.connect ("change-orient", self.__change_orient_cb)
        applet.show ()

        self.__set_plugin (self.__manager.active)
        self.__manager.connect ("notify::active", self.__active_changed_cb)
        self.__manager.connect ("notify::preferred",
                                lambda manager, pspec: self.__set_preferred (manager.preferred))

    def __make_transparent (self, widget):
        widget.connect ("realize", lambda widget: widget.window.set_back_pixmap (None, True))

    def __set_widget_tooltip (self, widget, text):
        if USE_GTK_TOOLTIP:
	    widget.set_tooltip_text (text)
	else:
	    self.__tooltips.set_tip (widget, text)

    def __query_tooltip_cb (self, widget, x, y, kbd, tooltip):
        # Yes, a new Guts each time, since GTK destroys it along with the tooltip itself.
        # First make sure the old one disconnects its handlers.
        self.__song_tip_guts.set_plugin (None)
        self.__song_tip_guts = SongTooltipGuts ()
        self.__song_tip_guts.show_all ()
        self.__song_tip_guts.set_plugin (self.__plugin)
        tooltip.set_custom (self.__song_tip_guts)
        return True

    def __set_plugin (self, plugin):
        for handler in self.__plugin_handlers:
            self.__plugin.disconnect (handler)

        self.__plugin = plugin
        if plugin is not None:
            self.__plugin_handlers = [
                plugin.connect ("notify::playing", self.__playing_changed_cb),
                plugin.connect ("notify::elapsed", self.__elapsed_changed_cb),
                plugin.connect ("notify::min_rating", self.__min_rating_changed_cb),
                plugin.connect ("notify::max_rating", self.__max_rating_changed_cb),
                plugin.connect ("notify::rating", self.__rating_changed_cb),
            ]
            self.__playing_changed_cb (plugin, None)
            self.__elapsed_changed_cb (plugin, None)
            self.__min_rating_changed_cb (plugin, None)
            self.__max_rating_changed_cb (plugin, None)
            self.__rating_changed_cb (plugin, None)
        else:
            self.__plugin_handlers = []
            self.__rating.interactive = False

        self.__scroller.props.plugin = plugin
        self.__song_tip_guts.set_plugin (plugin)
        if self.__notify is not None:
            self.__notify.set_plugin (plugin)
        self.__update_scroller_visibility ()
        self.__update_rating_visibility ()
        self.__update_time_visibility ()
        self.__update_controls_visibility ()
        self.__update_launch_visibility ()


    def __active_changed_cb (self, manager, pspec):
        self.__set_plugin (manager.active)


    def __set_preferred (self, preferred):
        if preferred is not None:
            if preferred.icon_name is not None:
                self.__launch.set_icon_name (preferred.icon_name)
            else:
                self.__launch.set_stock_id ("music-applet")
	    self.__set_widget_tooltip (self.__launch, _("Launch %s") % preferred.player_name)
        else:
            self.__launch.set_stock_id ("music-applet")
	    self.__set_widget_tooltip (self.__launch, _("Launch preferred player"))


    def __update_scroller_visibility (self, conf_value=None):
        full_key = self.__conf.resolve_key ("show_song_information")
        if conf_value is None:
            conf_value = self.__conf.client.get_bool (full_key)
        conf_value = self.__assure_not_invisible (full_key, conf_value)
        self.__scroller.props.visible = self.__plugin is not None and conf_value


    def __update_rating_visibility (self, conf_value=None):
        full_key = self.__conf.resolve_key ("show_rating")
        if conf_value is None:
            conf_value = self.__conf.client.get_bool (full_key)
        conf_value = self.__assure_not_invisible (full_key, conf_value)
        self.__rating.set_property ("visible", self.__plugin is not None and conf_value)


    def __update_time_visibility (self, conf_value=None):
        full_key = self.__conf.resolve_key ("show_time")
        if conf_value is None:
            conf_value = self.__conf.client.get_bool (full_key)
        conf_value = self.__assure_not_invisible (full_key, conf_value)
        self.__time.set_property ("visible", self.__plugin is not None and conf_value)


    def __update_controls_visibility (self, conf_value=None):
        full_key = self.__conf.resolve_key ("show_controls")
        if conf_value is None:
            conf_value = self.__conf.client.get_bool (full_key)
        conf_value = self.__assure_not_invisible (full_key, conf_value)
        for widget in [self.__previous, self.__play, self.__next]:
            widget.set_property ("visible", self.__plugin is not None and conf_value)


    def __assure_not_invisible (self, latest_full_key, conf_value):
        visible = False
        for key in ["show_song_information", "show_rating", "show_time", "show_controls"]:
            full_key = self.__conf.resolve_key (key)
            visible = visible or self.__conf.client.get_bool (full_key)
        if not visible:
            self.__conf.client.set_bool (latest_full_key, True)
            return True
        else:
            return conf_value


    def __update_launch_visibility (self):
        self.__launch.set_property ("visible", self.__plugin is None)


    def __change_background_cb (self, applet, type, color, pixmap):
        applet.set_style (None)
        applet.modify_style (gtk.RcStyle ())

        if type == gnomeapplet.NO_BACKGROUND:
            pass
        elif type == gnomeapplet.COLOR_BACKGROUND:
            applet.modify_bg (gtk.STATE_NORMAL, color)
        elif type == gnomeapplet.PIXMAP_BACKGROUND:
            style = applet.get_style ().copy ()
            style.bg_pixmap[gtk.STATE_NORMAL] = pixmap
            applet.set_style (style)

        for widget in [self.__scroller, self.__time, self.__rating]:
            widget.queue_draw ()


    def __change_orient_cb (self, applet, orient):
        widgets = [self.__scroller, self.__rating, self.__time, self.__previous, self.__play, self.__next, self.__launch]
        if self.__all is not None:
            for widget in widgets:
                self.__all.remove (widget)
            applet.remove (self.__all)

        if orient == gnomeapplet.ORIENT_UP or orient == gnomeapplet.ORIENT_DOWN:
            self.__all = gtk.HBox ()
        else:
            self.__all = gtk.VBox ()
        self.__make_transparent (self.__all)

        self.__all.pack_start (self.__scroller, padding=3)
        self.__all.pack_start (self.__rating)
        self.__all.pack_start (self.__time, padding=3)
        self.__all.pack_start (self.__previous)
        self.__all.pack_start (self.__play)
        self.__all.pack_start (self.__next)
        self.__all.pack_start (self.__launch)

        if orient == gnomeapplet.ORIENT_RIGHT:
            self.__time.set_angle (90)
        elif orient == gnomeapplet.ORIENT_LEFT:
            self.__time.set_angle (270)
        else:
            self.__time.set_angle (0)
        self.__scroller.props.orientation = orient

        self.__rating.vertical = (orient == gnomeapplet.ORIENT_LEFT or orient == gnomeapplet.ORIENT_RIGHT)
        for button in [self.__previous, self.__play, self.__next, self.__launch]:
            button.set_orientation (orient)

        applet.add (self.__all)
        self.__all.show ()


    ##################################################################
    #
    # Playback controls
    #
    ##################################################################

    def __previous_clicked_cb (self, button):
        self.__plugin.previous ()


    def __play_clicked_cb (self, button):
        self.__plugin.toggle_playback ()


    def __next_clicked_cb (self, button):
        self.__plugin.next ()


    def __launch_clicked_cb (self, button):
        if self.__manager.preferred is not None:
            self.__manager.preferred.launch ()
        else:
            self.__plugins_cb (None, None)


    def __rate_song_cb (self, rating, pspec):
        if rating.rating >= 0 and rating.rating != self.__plugin.rating and "rate_song" in dir (self.__plugin):
            self.__plugin.rate_song (rating.rating)


    ##################################################################
    #
    # Playback status
    #
    ##################################################################

    def __playing_changed_cb (self, plugin, pspec):
        if plugin.playing:
            self.__play.set_stock_id (gtk.STOCK_MEDIA_PAUSE)
	    self.__set_widget_tooltip (self.__play, _("Pause playback"))
        else:
            self.__play.set_stock_id (gtk.STOCK_MEDIA_PLAY)
	    self.__set_widget_tooltip (self.__play, _("Start playing"))


    def __elapsed_changed_cb (self, plugin, pspec):
        if plugin.elapsed < 0:
            self.__time.set_text ("--:--")
        else:
            self.__time.set_text (format_elapsed (plugin.elapsed))


    def __min_rating_changed_cb (self, plugin, pspec):
        self.__rating.minimum = plugin.min_rating


    def __max_rating_changed_cb (self, plugin, pspec):
        self.__rating.maximum = plugin.max_rating


    def __rating_changed_cb (self, plugin, pspec):
        self.__rating.rating = plugin.rating
        if "rate_song" not in dir (plugin):
            self.__set_widget_tooltip (self.__rating, _("%s does not support rating songs") %
                                                    self.__plugin.player_name)
            self.__rating.interactive = False
        elif plugin.rating < 0:
            self.__set_widget_tooltip (self.__rating, _("No song to rate"))
            self.__rating.interactive = False
        else:
            self.__set_widget_tooltip (self.__rating, _("Rate current song"))
            self.__rating.interactive = True


    ##################################################################
    #
    # Context menu
    #
    ##################################################################

    def __preferences_cb (self, component, verb):
        if self.__prefs_dialog is None:
            self.__prefs_dialog = musicapplet.conf.create_preferences_dialog (self.__conf)
            self.__prefs_dialog.connect ("destroy", self.__prefs_destroy_cb)
            self.__prefs_dialog.show ()
        else:
            self.__prefs_dialog.present ()


    def __prefs_destroy_cb (self, dialog):
        self.__prefs_dialog = None


    def __plugins_cb (self, component, verb):
        if self.__plugins_dialog is None:
            self.__plugins_dialog = musicapplet.manager.create_plugins_dialog (self.__manager, self.__conf)
            self.__plugins_dialog.connect ("destroy", self.__plugins_destroy_cb)
            self.__plugins_dialog.show ()
        else:
            self.__plugins_dialog.present ()


    def __plugins_destroy_cb (self, dialog):
        self.__plugins_dialog = None


    def __about_cb (self, component, verb):
        if self.__about_dialog is None:
            self.__about_dialog = gtk.AboutDialog ()
            self.__about_dialog.set_name (_("Music Applet"))
            self.__about_dialog.set_version (musicapplet.defs.VERSION)
            self.__about_dialog.set_copyright (_("(C) 2008 Paul Kuliniewicz"))
            self.__about_dialog.set_comments (_("Control your favorite music player from a GNOME panel."))
            self.__about_dialog.set_website ("http://www.kuliniewicz.org/music-applet/")
            self.__about_dialog.set_logo (self.__about_dialog.render_icon (STOCK_ICON, gtk.ICON_SIZE_LARGE_TOOLBAR))
            self.__about_dialog.set_authors ([
                "Paul Kuliniewicz <paul@kuliniewicz.org>",
            ])
            self.__about_dialog.set_translator_credits (_("translator-credits"))
            self.__about_dialog.connect ("response", lambda dialog, response: dialog.hide ())
            self.__about_dialog.connect ("destroy", self.__about_destroy_cb)
            self.__about_dialog.show ()
        else:
            self.__about_dialog.present ()


    def __about_destroy_cb (self, dialog):
        self.__about_dialog = None


class SongTooltip (musicapplet.widgets.FancyTooltips):
    """
    Old-style tooltip using a custom GtkTooltips subclass.
    """
    def __init__ (self, guts):
    	musicapplet.widgets.FancyTooltips.__init__ (self)
        self.__guts = guts
	self.__guts.show_all ()
	self.set_content (self.__guts)

    def set_plugin (self, plugin):
    	self.__guts.set_plugin (plugin)


class SongTooltipGuts (gtk.HBox):
    """
    Contents of a tooltip that displays information about the current song.
    """

    ##################################################################
    #
    # Construction and setup
    #
    ##################################################################

    def __init__ (self):
	gtk.HBox.__init__(self, False, 6)
        self.__handlers = []

        self.__title = gtk.Label ()
        self.__artist = gtk.Label ()
        self.__album = gtk.Label ()
        self.__time = gtk.Label ()
        self.__art = gtk.Image ()

        vbox = gtk.VBox ()
        for label in [self.__title, self.__artist, self.__album, self.__time]:
            label.set_alignment (0.0, 0.5)
            vbox.add (label)

	self.add (self.__art)
	self.add (vbox)


    def set_plugin (self, plugin):
        for handler in self.__handlers:
            self.__plugin.disconnect (handler)
        self.__plugin = plugin

        if self.__plugin is not None:
            self.__handlers = [
                self.__plugin.connect ("notify::title", lambda plugin, pspec: self.__set_title (plugin.title)),
                self.__plugin.connect ("notify::artist", lambda plugin, pspec: self.__set_artist (plugin.artist)),
                self.__plugin.connect ("notify::album", lambda plugin, pspec: self.__set_album (plugin.album)),
                self.__plugin.connect ("notify::elapsed", lambda plugin, pspec: self.__set_time (plugin.elapsed, plugin.duration)),
                self.__plugin.connect ("notify::duration", lambda plugin, pspec: self.__set_time (plugin.elapsed, plugin.duration)),
                self.__plugin.connect ("notify::art", lambda plugin, pspec: self.__set_art (plugin.art)),
            ]

            self.__set_title (self.__plugin.title)
            self.__set_artist (self.__plugin.artist)
            self.__set_album (self.__plugin.album)
            self.__set_time (self.__plugin.elapsed, self.__plugin.duration)
            self.__set_art (self.__plugin.art)

        else:
            self.__handlers = []
            self.__set_title (None)


    ##################################################################
    #
    # Title
    #
    ##################################################################

    def __set_title (self, title):
        if title is not None:
            self.__title.set_markup ("<big><b>%s</b></big>" % saxutils.escape (title))
        else:
            self.__title.set_text (_("Not playing"))


    def __set_artist (self, artist):
        if artist is not None:
            self.__artist.set_markup (_("<i>by</i> %s") % saxutils.escape (artist))
            self.__artist.show ()
        else:
            self.__artist.hide ()


    def __set_album (self, album):
        if album is not None:
            self.__album.set_markup (_("<i>from</i> %s") % saxutils.escape (album))
            self.__album.show ()
        else:
            self.__album.hide ()


    def __set_time (self, elapsed, duration):
        if duration > 0:
            self.__time.set_text (format_elapsed_duration (elapsed, duration))
            self.__time.show ()
        elif elapsed > 0:
            self.__time.set_text (format_elapsed (elapsed))
            self.__time.show ()
        else:
            self.__time.hide ()


    def __set_art (self, art):
        if art is not None:
            self.__art.set_from_pixbuf (scale_pixbuf (art))
            self.__art.show ()
        else:
            self.__art.set_from_pixbuf (None)
            self.__art.hide ()


class Rating (gtk.EventBox):
    """
    Displays a rating between zero and five stars.
    """

    MAJOR_OFFSET = 4
    MINOR_OFFSET = 2

    __gproperties__ = {
        "rating" : (gobject.TYPE_DOUBLE,
                    "rating",
                    "The rating, measured in stars",
                    -1.0, 10.0, -1.0,        # -1.0 means ratings are disabled
                    gobject.PARAM_READWRITE),

        "minimum" : (gobject.TYPE_DOUBLE,
                     "minimum",
                     "The minimum allowable rating",
                     -1.0, 1.0, -1.0,
                     gobject.PARAM_READWRITE),

        "maximum" : (gobject.TYPE_DOUBLE,
                     "maximum",
                     "The maximum allowable rating",
                     -1.0, 10.0, -1.0,
                     gobject.PARAM_READWRITE),

        "vertical" : (gobject.TYPE_BOOLEAN,
                      "vertical",
                      "Whether the rating stars should be lined up vertically",
                      False,
                      gobject.PARAM_READWRITE),

        "interactive" : (gobject.TYPE_BOOLEAN,
                         "interactive",
                         "Whether the widget should allow the user to set a new rating",
                         False,
                         gobject.PARAM_READWRITE)
    }

    rating = musicapplet.util.make_property ("rating")
    minimum = musicapplet.util.make_property ("minimum")
    maximum = musicapplet.util.make_property ("maximum")
    vertical = musicapplet.util.make_property ("vertical")
    interactive = musicapplet.util.make_property ("interactive")


    def __init__ (self):
        gtk.EventBox.__init__ (self)
        self.add_events (gdk.POINTER_MOTION_MASK)

        self.__set_star_stock = self.render_icon (STOCK_SET_STAR, gtk.ICON_SIZE_MENU)
        self.__unset_star_stock = self.render_icon (STOCK_UNSET_STAR, gtk.ICON_SIZE_MENU)
        self.__set_star = self.__set_star_stock
        self.__unset_star = self.__unset_star_stock
        self.__star_width = max (self.__set_star.get_width (), self.__unset_star.get_width ())
        self.__star_height = max (self.__set_star.get_height (), self.__unset_star.get_height ())

        self.__rating = -1.0
        self.__minimum = -1.0
        self.__maximum = -1.0
        self.__vertical = False
        self.__mouseover = None
        self.__interactive = False


    def do_style_set (self, old_style):
        gtk.EventBox.do_style_set (self, old_style)

        # render_icon causes this to be called, so...
        if "_Rating__set_star_stock" in dir (self) and "_Rating__unset_star_stock" in dir (self):
            color = self.get_style ().fg[gtk.STATE_NORMAL]
            self.__set_star   = self.create_colorized_pixbuf (self.__set_star_stock, color)
            self.__unset_star = self.create_colorized_pixbuf (self.__unset_star_stock, color)
            self.queue_draw ()


    def do_get_property (self, property):
        if property.name == "rating":
            return self.__rating
        elif property.name == "minimum":
            return self.__minimum
        elif property.name == "maximum":
            return self.__maximum
        elif property.name == "vertical":
            return self.__vertical
        elif property.name == "interactive":
            return self.__interactive
        else:
            raise AttributeError, "unknown property '%s'" % property.name


    def do_set_property (self, property, value):
        if property.name == "rating":
            self.__rating = value
            self.queue_draw ()
        elif property.name == "minimum":
            self.__minimum = value
            self.queue_resize ()
        elif property.name == "maximum":
            self.__maximum = value
            self.queue_resize ()
        elif property.name == "vertical":
            self.__vertical = value
            self.queue_resize ()
        elif property.name == "interactive":
            self.__interactive = value
            self.queue_draw ()
        else:
            raise AttributeError, "unknown property '%s'" % property.name


    def do_size_request (self, requisition):
        if self.__minimum < 0.0:
            requisition.width = 0
            requisition.height = 0
        elif self.__vertical:
            requisition.width = self.__star_width + self.MINOR_OFFSET * 2
            requisition.height = self.__star_height * self.__maximum + self.MAJOR_OFFSET * 2
        else:
            requisition.width = self.__star_width * self.__maximum + self.MAJOR_OFFSET * 2
            requisition.height = self.__star_height + self.MINOR_OFFSET * 2


    def do_button_press_event (self, event):
        if self.__rating >= 0.0 and event.button == 1 and self.__mouseover is not None:
            if self.__interactive:
                self.rating = self.__mouseover
            return True
        else:
            return False


    def do_enter_notify_event (self, event):
        self.__update_mouseover (event)
        return False


    def do_motion_notify_event (self, event):
        self.__update_mouseover (event)
        return False

    def do_leave_notify_event (self, event):
        self.__mouseover = None
        self.queue_draw ()
        return False


    def do_expose_event (self, event):
        rtl = (self.get_direction () == gtk.TEXT_DIR_RTL)
        if self.__mouseover is not None and self.__interactive:
            stars = self.__mouseover
        else:
            stars = self.__rating

        for pos in range (int (self.__maximum)):
            if pos < stars:
                pixbuf = self.__set_star
            else:
                pixbuf = self.__unset_star

            if self.__vertical:
                x_offset = (self.allocation.width - self.__star_width) / 2
                y_offset = self.MAJOR_OFFSET + (self.__maximum - pos - 1) * self.__star_height
            elif rtl:
                x_offset = self.MAJOR_OFFSET + (self.__maximum - pos - 1) * self.__star_width
                y_offset = (self.allocation.height - self.__star_height) / 2
            else:
                x_offset = self.MAJOR_OFFSET + pos * self.__star_width
                y_offset = (self.allocation.height - self.__star_height) / 2

            self.window.draw_pixbuf (None, pixbuf,
                                     0, 0,
                                     x_offset, y_offset,
                                     self.__star_width, self.__star_height,
                                     gdk.RGB_DITHER_NORMAL, 0, 0)
        return True


    def __update_mouseover (self, event):
        mouseover = None
        if event.x >= 0 and event.x <= self.allocation.width and event.y >= 0 and event.y <= self.allocation.height:
            if self.__vertical:
                mouseover = self.__maximum - ((event.y - self.MAJOR_OFFSET) / self.__star_height) + 1
            else:
                if event.x <= self.MAJOR_OFFSET:
                    mouseover = 0
                else:
                    mouseover = ((event.x - self.MAJOR_OFFSET) / self.__star_width) + 1

                if self.get_direction () == gtk.TEXT_DIR_RTL:
                    mouseover = self.__maximum - mouseover + 1

            mouseover = int (mouseover)

            if mouseover > self.__maximum:
                mouseover = self.__maximum
            if mouseover < self.__minimum:
                mouseover = self.__minimum

        if self.__mouseover != mouseover:
            self.__mouseover = mouseover
            self.queue_draw ()


    def create_colorized_pixbuf (self, stock_pixbuf, color):
        pixbuf = stock_pixbuf.copy ()
        red_scale = color.red / 65535.0
        green_scale = color.green / 65535.0
        blue_scale = color.blue / 65535.0
        for row in pixbuf.get_pixels_array ():
            for pixel in row:
                # Yay API changes!
                if str (type (pixel[0])) == "<type 'array'>":
                    pixel[0][0] *= red_scale
                    pixel[1][0] *= green_scale
                    pixel[2][0] *= blue_scale
                else:
                    pixel[0] *= red_scale
                    pixel[1] *= green_scale
                    pixel[2] *= blue_scale
        return pixbuf



class NotifyPopup (gobject.GObject):
    """
    Interface between the applet and a libnotify popup, if pynotify is
    available.
    """

    def __init__ (self, applet, conf):
        gobject.GObject.__init__ (self)

        import pynotify
        pynotify.init ("music-applet")

        self.__notify = None
        self.__plugin = None
        self.__applet = applet
        self.__conf = conf
        self.__key = conf.resolve_key ("show_notification")
        self.__handlers = []
        self.__show_source = None

        conf.client.notify_add (self.__key, self.__notification_config_changed_cb)


    def set_plugin (self, plugin):
        for handler in self.__handlers:
            self.__plugin.disconnect (handler)
        self.__plugin = plugin

        if self.__plugin is not None:
            self.__handlers = [
                self.__plugin.connect ("notify::title", lambda plugin, pspec: self.__queue_show ()),
                self.__plugin.connect ("notify::artist", lambda plugin, pspec: self.__queue_show ()),
                self.__plugin.connect ("notify::album", lambda plugin, pspec: self.__queue_show ()),
                self.__plugin.connect ("notify::art", lambda plugin, pspec: self.__queue_show ()),
            ]
        else:
            self.__handlers = []

        self.__show ()


    def __show (self):
        """
        Show the notification popup, as long as there's a current song.
        """

        import pynotify

        if self.__notify is not None:
            self.__notify.close ()

        if self.__plugin is not None and self.__plugin.title is not None and self.__conf.client.get_bool (self.__key):
            body = []
            if self.__plugin.artist is not None:
                body.append (_("<i>by</i> %s") % saxutils.escape (self.__plugin.artist))
            if self.__plugin.album is not None:
                body.append (_("<i>from</i> %s") % saxutils.escape (self.__plugin.album))

            self.__notify = pynotify.Notification (saxutils.escape (self.__plugin.title), "\n".join (body), None, self.__applet)
            self.__notify.set_urgency (pynotify.URGENCY_LOW)
            if self.__plugin.art is not None:
                self.__notify.set_icon_from_pixbuf (scale_pixbuf (self.__plugin.art))

            self.__notify.show ()
        else:
            self.__notify = None

        self.__show_source = None
        return False


    def __queue_show (self):
        """
        Queue a display of the popup in the very near future.

        This is used to prevent repeatedly creating and destroying the popup
        as the plugin populates its metadata properties individually, or at
        least make it much less likely.
        """

        if self.__show_source is None:
            self.__show_source = gobject.timeout_add (500, self.__show)


    def __notification_config_changed_cb (self, client, id, entry, unused):
        if entry.value is not None and not entry.value.get_bool () and self.__notify is not None:
            self.__notify.close ()
            self.__notify = None


def format_elapsed (elapsed):
    """
    Convert an elapsed time into a human-readable format.
    """

    (hours, minutes, seconds) = split_time (elapsed)
    if hours > 0:
        # To translators: HOURS:MINUTES:SECONDS
        return _("%d:%02d:%02d") % (hours, minutes, seconds)
    else:
        # To translators: MINUTES:SECONDS
        return _("%d:%02d") % (minutes, seconds)

def format_elapsed_duration (elapsed, duration):
    """
    Convert an elapsed time and total duration into a human-readable format.
    """

    (elapsed_hours, elapsed_minutes, elapsed_seconds) = split_time (elapsed)
    (duration_hours, duration_minutes, duration_seconds) = split_time (duration)

    if elapsed_hours > 0 or duration_hours > 0:
        # To translators: ELAPSED of DURATION, using HOURS:MINUTES:SECONDS
        return _("%d:%02d:%02d of %d:%02d:%02d") % (elapsed_hours, elapsed_minutes, elapsed_seconds,
                                                    duration_hours, duration_minutes, duration_seconds)
    else:
        # To translators: ELAPSED of DURATION, using MINUTES:SECONDS
        return _("%d:%02d of %d:%02d") % (elapsed_minutes, elapsed_seconds,
                                          duration_minutes, duration_seconds)

def split_time (time):
    """
    Split a time value into its constituent hours, minutes, and seconds.
    """

    seconds = time % 60
    minutes = (time / 60) % 60
    hours = time / 3600
    return (hours, minutes, seconds)


def register_stock_icons ():
    """
    Register the stock icons used by the various classes in the module.
    """

    factory = gtk.IconFactory ()
    factory.add_default ()

    for stock_id in [STOCK_ICON, STOCK_SET_STAR, STOCK_UNSET_STAR]:
        pixbuf = gdk.pixbuf_new_from_file (os.path.join (musicapplet.defs.PKG_DATA_DIR, "%s.png" % stock_id))
        icon_set = gtk.IconSet (pixbuf)
        factory.add (stock_id, icon_set)


def augment_icon_path ():
    """
    Augment the icon search path with where Music Applet places its own,
    to assure they appear even if the applet is installed to a nonstandard
    path.
    """

    theme = gtk.icon_theme_get_default ()
    theme.append_search_path (os.path.join (musicapplet.defs.DATA_DIR, "icons"))


def scale_pixbuf (pixbuf):
    """
    Created a scaled version of a pixbuf which is more suitable for disaplying
    in a tooltip-like window.
    """

    # This could be bad if the art is ridiculously wide, but that's pretty unlikely.
    return pixbuf.scale_simple (64 * pixbuf.get_width () / pixbuf.get_height (), 64, gdk.INTERP_BILINEAR)



gobject.type_register (NotifyPopup)
gobject.type_register (Rating)
register_stock_icons ()
augment_icon_path ()



# Music Applet
# Copyright (C) 2007-2008 Paul Kuliniewicz <paul@kuliniewicz.org>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02111-1301, USA.

import musicapplet.defs
import musicapplet.player

import gobject
from gtk import gdk


INFO = {
        "display-name": "Banshee (pre-1.0)",
        "internal-name": "Banshee",
        "version": musicapplet.defs.VERSION,
        "icon-name": "music-player-banshee",
        "author": "Paul Kuliniewicz <paul@kuliniewicz.org>",
        "copyright": "(C) 2007-2008 Paul Kuliniewicz",
        "website": "http://www.kuliniewicz.org/music-applet/",
}


class BansheePlugin (musicapplet.player.DBusPlugin):
    """
    Music Applet plugin to interface with Banshee (pre-1.0 versions).
    """


    ##################################################################
    #
    # Creation and overall management
    #
    ##################################################################


    def __init__ (self, conf):
        musicapplet.player.DBusPlugin.__init__ (self, conf, "Banshee", "music-player-banshee",
                                                "org.gnome.Banshee")

        self.min_rating = 0.0
        self.max_rating = 5.0

        self.__uri = None
        self.__poll_source = None

        self.__player = None


    ##################################################################
    #
    # Playback control
    #
    ##################################################################


    def toggle_playback (self):
        self.__player.TogglePlaying (reply_handler=lambda: None,
                                     error_handler=self._report_dbus_error)


    def previous (self):
        self.__player.Previous (reply_handler=lambda: None,
                                error_handler=self._report_dbus_error)


    def next (self):
        self.__player.Next (reply_handler=lambda: None,
                            error_handler=self._report_dbus_error)


    ##################################################################
    #
    # URI
    #
    ##################################################################


    def __set_uri (self, uri):
        if uri != self.__uri:
            if uri is not None and uri != "":
                self.__uri = uri
                self.__player.GetPlayingTitle (reply_handler=self.__set_title,
                                               error_handler=self._report_dbus_error)

                self.__player.GetPlayingArtist (reply_handler=self.__set_artist,
                                                error_handler=self._report_dbus_error)

                self.__player.GetPlayingAlbum (reply_handler=self.__set_album,
                                               error_handler=self._report_dbus_error)

                self.__player.GetPlayingDuration (reply_handler=self.__set_duration,
                                                  error_handler=self._report_dbus_error)

                self.__player.GetPlayingRating (reply_handler=self.__set_rating,
                                                error_handler=self._report_dbus_error)

                self.__player.GetPlayingCoverUri (reply_handler=self.__set_art,
                                                  error_handler=self._report_dbus_error)
            else:
                self._set_no_song ()


    ##################################################################
    #
    # Title
    #
    ##################################################################


    def __set_title (self, title):
        self.title = title


    ##################################################################
    #
    # Artist
    #
    ##################################################################


    def __set_artist (self, artist):
        self.artist = artist


    ##################################################################
    #
    # Album
    #
    ##################################################################


    def __set_album (self, album):
        self.album = album


    ##################################################################
    #
    # Duration
    #
    ##################################################################


    def __set_duration (self, duration):
        # XXX: Does Banshee still sometimes return a bogus duration?
        self.duration = duration


    ##################################################################
    #
    # Rating
    #
    ##################################################################


    def __set_rating (self, rating):
        self.rating = rating


    def rate_song (self, rating):
        self.__player.SetPlayingRating (rating,
                                        reply_handler=lambda old: self.set_property ("rating", rating),
                                        error_handler=self._report_dbus_error)


    ##################################################################
    #
    # Playing
    #
    ##################################################################


    def __set_playing (self, playing):
        self.playing = (playing == 1)


    ##################################################################
    #
    # Elapsed
    #
    ##################################################################


    def __set_elapsed (self, elapsed):
        self.elapsed = elapsed


    ##################################################################
    #
    # Art
    #
    ##################################################################


    def __set_art (self, art_uri):
        if art_uri is not None and len (art_uri) > 0:
            self.art = gdk.pixbuf_new_from_file (art_uri)
        else:
            self.art = None


    ##################################################################
    #
    # Connection management
    #
    ##################################################################


    def _connect (self):
        if self.__player is not None:
            return

        import dbus
        bus = dbus.SessionBus ()

        proxy = bus.get_object ("org.gnome.Banshee", "/org/gnome/Banshee/Player")
        self.__player = dbus.Interface (proxy, "org.gnome.Banshee.Core")

        self.__poll ()
        self.connected = True
        self.__poll_source = gobject.timeout_add (1000, self.__poll)


    def _disconnect (self):
        self.__player = None

        if self.__poll_source is not None:
            gobject.source_remove (self.__poll_source)
            self.__poll_source = None

        self._set_no_song ()
        self.connected = False


    def __poll (self):
        self.__player.GetPlayingUri (reply_handler=self.__set_uri,
                                     error_handler=self._report_dbus_error)

        self.__player.GetPlayingStatus (reply_handler=self.__set_playing,
                                        error_handler=self._report_dbus_error)

        self.__player.GetPlayingPosition (reply_handler=self.__set_elapsed,
                                          error_handler=self._report_dbus_error)

        # XXX: Need to watch for http://bugzilla.gnome.org/show_bug.cgi?id=344774 ?

        return True


def create_instance (conf):
    return BansheePlugin (conf)



# Music Applet
# Copyright (C) 2007 Paul Kuliniewicz <paul@kuliniewicz.org>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02111-1301, USA.

import musicapplet.defs
import musicapplet.player

from gettext import gettext as _
import gobject
import gtk
from gtk import glade

import os
import Queue
import random
import socket
import sys              # DEBUG ONLY
import threading
import time


INFO = {
        "name": "MPD",
        "version": musicapplet.defs.VERSION,
        "icon-name": "music-applet-mpd",
        "author": "Paul Kuliniewicz <paul@kuliniewicz.org>",
        "copyright": "(C) 2007 Paul Kuliniewicz",
        "website": "http://www.kuliniewicz.org/music-applet/",
}


class MPDPlugin (musicapplet.player.PluginBase):


    def __init__ (self, conf):
        musicapplet.player.PluginBase.__init__ (self, conf, INFO["name"], INFO["icon-name"])

        import mpdclient2

        self.__thread = None


    def start (self):
        if self.__thread is None:
            self.__thread = MPDThread (self)
            self.__thread.start ()


    def stop (self):
        if self.__thread is not None:
            self.__thread.cancel = True
            self.__thread = None

        self._set_no_song ()
        self.connected = False


    def toggle_playback (self):
        self.__thread.queue.put (("toggle_playback",))


    def previous (self):
        self.__thread.queue.put (("previous",))


    def next (self):
        self.__thread.queue.put (("next",))


    # def rate_song (self, rating):
    #     pass


    def create_config_dialog (self):
        xml = glade.XML (os.path.join (musicapplet.defs.PKG_DATA_DIR, "mpd.glade"))
        dialog = xml.get_widget ("mpd-dialog")

        full_key = self._conf.resolve_plugin_key (self, "host")
        self._conf.bind_string (full_key, xml.get_widget ("host"), "text")

        full_key = self._conf.resolve_plugin_key (self, "port")
        self._conf.bind_int (full_key, xml.get_widget ("port"), "value")

        full_key = self._conf.resolve_plugin_key (self, "command")
        self._conf.bind_string (full_key, xml.get_widget ("command"), "text")

        xml.get_widget ("browse").connect ("clicked",
                    lambda button: self._browse_for_command (dialog, _("%s Launcher") % "MPD"))
        dialog.set_title (_("%s Plugin") % "MPD")
        dialog.set_default_response (gtk.RESPONSE_CLOSE)
        dialog.connect ("response", lambda dialog, response: dialog.hide ())
        return dialog


    def launch (self):
        self.__thread.queue.put (("launch",))


    def _show_password_dialog (self):
        xml = glade.XML (os.path.join (musicapplet.defs.PKG_DATA_DIR, "password.glade"))
        dialog = xml.get_widget ("password-dialog")

        xml.get_widget ("message").set_text (_("A password is needed to connect to the MPD server at %s:%d.") %
                                             (self.__thread.host, self.__thread.port))

        dialog.set_default_response (gtk.RESPONSE_OK)
        dialog.connect ("response", self.__password_response_cb, xml)
        dialog.show ()

        return False


    def __password_response_cb (self, dialog, response, xml):
        if response == gtk.RESPONSE_OK:
            self.__thread.queue.put (("password",
                                      xml.get_widget ("password").get_text (),
                                      xml.get_widget ("remember").get_active ()))
        else:
            self.__thread.queue.put (("cancel",))
        dialog.destroy ()


class MPDThread (threading.Thread):
    """
    Separate thread that does all the actual interaction with MPD, since the
    MPD client library uses blocking I/O, which can cause unpredictable and
    large delays when running MPD over a network.
    """


    MIN_INITIAL_DELAY = 8       # seconds
    MAX_INITIAL_DELAY = 60      # seconds
    MAX_DELAY = 600             # seconds


    def __init__ (self, plugin):
        threading.Thread.__init__ (self, name="MPD Thread")
        self.setDaemon (True)
        self.queue = Queue.Queue ()
        self.cancel = False

        # These tell the password dialog where the server is; they aren't
        # actually used for connecting; mpdclient2 handles the defaults there.
        self.host = None
        self.port = None

        self.__plugin = plugin
        self.__mpd = None
        self.__delay = 0
        self.__autoconnect = True
        self.__authenticated = False


    def run (self):
        while not self.cancel:
            sys.stdout.flush ()
            if self.__mpd is None and self.__autoconnect:
                self.__try_connect ()

            if self.__authenticated:
                self.__poll ()
            elif self.__mpd is None and self.__autoconnect:
                if self.__delay <= 0:
                    self.__delay = random.randint (self.MIN_INITIAL_DELAY, self.MAX_INITIAL_DELAY)
                else:
                    self.__delay = min (self.__delay * 2, self.MAX_DELAY)
            else:
                self.__delay = None

            if not self.cancel:
                self.__process_queue (self.__delay)

        sys.stdout.flush ()


    def __try_connect (self):
        import mpdclient2
        try:
            full_key = self.__plugin._conf.resolve_plugin_key (self.__plugin, "host")
            host = self.__plugin._conf.client.get_string (full_key)
            if host == "":
                host = None
                self.host = os.environ.get ("MPD_HOST", "localhost")
                if "@" in self.host:
                    self.host = (self.host.split ("@", 1))[1]
            else:
                self.host = host

            full_key = self.__plugin._conf.resolve_plugin_key (self.__plugin, "port")
            port = self.__plugin._conf.client.get_int (full_key)
            if port <= 0:
                port = None
                self.port = int (os.environ.get ("MPD_PORT", 6600))
            else:
                self.port = port

            sys.stdout.flush ()
            self.__mpd = mpdclient2.connect (host=host, port=port)

            if not self.__authenticate ():
                try:
                    import gnomekeyring
                    keyring = gnomekeyring.get_default_keyring_sync ()
                    found = gnomekeyring.find_items_sync (gnomekeyring.ITEM_NETWORK_PASSWORD,
                                                          { "server":self.host, "port":self.port, "protocol":"mpd" })
                    for item in found:
                        if self.__authenticate (item.secret, False):
                            break
                except ImportError:
                    print "MPD plugin: unable to check for saved passwords"
                except gnomekeyring.DeniedError:
                    print "MPD plugin: no passwords found"
                except AttributeError:
                    print "MPD plugin: password lookup requires gnomekeyring >= 2.18.0"

            if not self.__authenticated:
                gobject.idle_add (self.__plugin._show_password_dialog)

        except socket.error:
            pass


    def __authenticate (self, password=None, save=None):
        if password is not None:
            self.__mpd.password (password)

        needed_commands = ("status", "currentsong", "play", "previous", "next")
        if len ([c for c in self.__mpd.notcommands () if "command" in c and c.command in needed_commands]) > 0:
            return False

        if save:
            try:
                import gnomekeyring
                keyring = gnomekeyring.get_default_keyring_sync ()
                result = gnomekeyring.item_create_sync (keyring,
                                                        gnomekeyring.ITEM_NETWORK_PASSWORD,
                                                        "Password for MPD server at %s:%d" % (self.host, self.port),
                                                        { "server":self.host, "port":self.port, "protocol":"mpd" },
                                                        password,
                                                        True)
            except ImportError:
                print "MPD plugin: unable to save password; gnomekeyring not found"

        self.__delay = 1
        self.__authenticated = True
        gobject.idle_add (self.__set_connected, True)
        return True


    def __set_connected (self, connected):
        if not connected:
            self.__plugin._set_no_song ()
        self.__plugin.connected = connected
        return False


    def __poll (self):
        try:
            sys.stdout.flush ()
            status = self.__mpd.status ()
            if status.state != "stop":
                song = self.__mpd.currentsong ()
            else:
                song = None
            gobject.idle_add (self.__set_status, status, song)
        except EOFError:
            sys.stdout.flush ()
            self.__mpd = None
            self.__authenticated = False
            gobject.idle_add (self.__set_connected, False)


    def __set_status (self, status, song):
        self.__plugin.playing = (status.state == "play")
        if song is not None:
            if self.__plugin.title != song.title:
                self.__plugin.title = song.title

            if "artist" in song:
                if self.__plugin.artist != song.artist:
                    self.__plugin.artist = song.artist
            else:
                self.__plugin.artist = None

            if "album" in song:
                if self.__plugin.album != song.album:
                    self.__plugin.album = song.album
            else:
                self.__plugin.album = None

            self.__plugin.duration = int (song.time)
            self.__plugin.elapsed = int (status.time.split (":")[0])
        else:
            self.__plugin._set_no_song ()

        return False


    def __process_queue (self, delay=None):
        try:
            # Wait up to 'delay' seconds for a command.  Once a command arrives,
            # process the entire queue and finish, since the command probably
            # changed MPD's status somehow, and we want to fetch it ASAP.
            command = self.queue.get (True, delay)
            while True:
                if command[0] == "toggle_playback":
                    self.__mpd.pause (self.__plugin.playing)
                elif command[0] == "previous":
                    self.__mpd.previous ()
                elif command[0] == "next":
                    self.__mpd.next ()
                elif command[0] == "password":
                    if not self.__authenticate (command[1], command[2]):
                        gobject.idle_add (self.__plugin._show_password_dialog)
                elif command[0] == "cancel":
                    try:
                        self.__mpd.close ()
                    except EOFError:
                        # Always happens -- mpdclient2 doesn't handle close() gracefully
                        pass
                    self.__mpd = None
                    self.__authenticated = False
                    self.__autoconnect = False
                elif command[0] == "launch":
                    if self.__mpd is None and not self.__autoconnect:
                        # We stopped trying to connect, so maybe it's actually running...
                        self.__try_connect ()
                    if self.__mpd is None:
                        # Nope, need to launch it after all.
                        gobject.idle_add (lambda: musicapplet.player.PluginBase.launch (self.__plugin) and False)
                        self.__autoconnect = True
                        self.__delay = 1
                else:
                    print "MPD plugin: (internal) unknown command '%s'" % command[0]
                command = self.queue.get (False)
        except Queue.Empty:
            pass


    def __sleep (self, delay):
        # Since the timeout handlers in other plugins might make sleep() wake
        # up much earlier than we like....
        start = time.time ()
        while start + delay > time.time ():
            time.sleep (start + delay - time.time ())


def create_instance (conf):
    return MPDPlugin (conf)



# Music Applet
# Copyright (C) 2007 Paul Kuliniewicz <paul@kuliniewicz.org>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02111-1301, USA.

from gettext import gettext as _
import gobject
import gtk
from gtk import glade

import os

import musicapplet.defs
import musicapplet.player


INFO = {
        "name": "XMMS",
        "version": musicapplet.defs.VERSION,
        "icon-name": "xmms",
        "author": "Paul Kuliniewicz <paul@kuliniewicz.org>",
        "copyright": "(C) 2007 Paul Kuliniewicz",
        "website": "http://www.kuliniewicz.org/music-applet/",
}


class XMMSPlugin (musicapplet.player.PluginBase):
    """
    Music Applet plugin to interface with XMMS 1.x.
    """

    def __init__ (self, conf):
        musicapplet.player.PluginBase.__init__ (self, conf, INFO["name"], INFO["icon-name"])

        # Note that this file is named xmms1.py instead of xmms.py, to
        # avoid confusion with the system's xmms module.

        import xmms.control

        self.__poll_source = None

        # Since the session ID is used so frequently, and changes so rarely,
        # save a copy of its value in the class instead of querying GConf every
        # single time.

        full_key = conf.resolve_plugin_key (self, "session")
        self.__session = conf.client.get_int (full_key)
        conf.client.notify_add (full_key, self.__session_changed_cb)


    def __session_changed_cb (self, conf, id, entry, unused):
        if entry.value is not None:
            self.__session = entry.value.get_int ()


    def start (self):
        if self.__poll_source is None:
            self.__poll_source = gobject.timeout_add (1000, self.__poll)


    def stop (self):
        if self.__poll_source is not None:
            gobject.source_remove (self.__poll_source)
            self.__poll_source = None
            self._set_no_song ()
            self.connected = False


    def toggle_playback (self):
        import xmms.control
        xmms.control.play_pause (self.__session)


    def previous (self):
        import xmms.control
        xmms.control.playlist_prev (self.__session)


    def next (self):
        import xmms.control
        xmms.control.playlist_next (self.__session)


    def create_config_dialog (self):
        xml = glade.XML (os.path.join (musicapplet.defs.PKG_DATA_DIR, "xmms1.glade"))
        dialog = xml.get_widget ("xmms-dialog")

        full_key = self._conf.resolve_plugin_key (self, "command")
        self._conf.bind_string (full_key, xml.get_widget ("command"), "text")

        full_key = self._conf.resolve_plugin_key (self, "session")
        self._conf.bind_int (full_key, xml.get_widget ("session"), "value")

        xml.get_widget ("browse").connect ("clicked", lambda button: self._browse_for_command (dialog, "XMMS"))
        dialog.set_title (_("%s Plugin") % "XMMS")
        dialog.set_default_response (gtk.RESPONSE_CLOSE)
        dialog.connect ("response", lambda dialog, response: dialog.hide ())
        return dialog


    def __poll (self):
        import xmms.control

        # XMMS is pretty spartan in terms of its feature set....

        self.connected = xmms.control.is_running (self.__session)
        if self.connected:
            if xmms.control.is_playing (self.__session):
                self.playing = not xmms.control.is_paused (self.__session)
                index = xmms.control.get_playlist_pos (self.__session)
                title = xmms.control.get_playlist_title (index, self.__session)
                if self.title != title:
                    self.title = title
                self.duration = xmms.control.get_playlist_time (index, self.__session) / 1000
                self.elapsed = xmms.control.get_output_time (self.__session) / 1000
            else:
                self.playing = False
                self.title = None
                self.duration = -1
                self.elapsed = -1
        else:
            self._set_no_song ()

        return True


def create_instance (conf):
    return XMMSPlugin (conf)



# Music Applet
# Copyright (C) 2004-2006 Paul Kuliniewicz <paul@kuliniewicz.org>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02111-1301, USA.

"""
Various compile-time definitions.

Plugins will be most interested in:
    * VERSION: the version of Music Applet they are running in
    * PKG_DATA_DIR: the directory where Music Applet's data files are
"""

import os

DATA_DIR = "/usr/share"
LIB_DIR = "/usr/lib"
VERSION = "2.4.2"
PACKAGE = "music-applet"
PYTHON_DIR = "/usr/lib/python2.5/site-packages"
PKG_DATA_DIR = "/usr/share/music-applet"



# Music Applet
# Copyright (C) 2008 Paul Kuliniewicz <paul@kuliniewicz.org>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02111-1301, USA.

import musicapplet.defs
import musicapplet.player

import gobject
from gtk import gdk

import os


INFO = {
        "display-name": "Banshee",
        "internal-name": "Banshee 1.0",
        "version": musicapplet.defs.VERSION,
        "icon-name": "media-player-banshee",
        "author": "Paul Kuliniewicz <paul@kuliniewicz.org>",
        "copyright": "(C) 2008 Paul Kuliniewicz",
        "website": "http://www.kuliniewicz.org/music-applet/",
}


class Banshee1Plugin (musicapplet.player.DBusPlugin):
    """
    Music Applet plugin to interface with Banshee (version 1.0 and later).
    """

    def __init__ (self, conf):
        musicapplet.player.DBusPlugin.__init__ (self,
                                                conf,
                                                { "display": INFO["display-name"], "internal": INFO["internal-name"] },
                                                INFO["icon-name"],
                                                "org.bansheeproject.Banshee")

        self.__handlers = []
        self.__poll_source = None
        self.__engine = None
        self.__controller = None

    def _connect (self):
        if self.__engine is not None and self.__controller is not None:
            return

        import dbus
        bus = dbus.SessionBus ()

        proxy = bus.get_object ("org.bansheeproject.Banshee", "/org/bansheeproject/Banshee/PlayerEngine")
        self.__engine = dbus.Interface (proxy, "org.bansheeproject.Banshee.PlayerEngine")

        proxy = bus.get_object ("org.bansheeproject.Banshee", "/org/bansheeproject/Banshee/PlaybackController")
        self.__controller = dbus.Interface (proxy, "org.bansheeproject.Banshee.PlaybackController")

        self.__handlers = [
            self.__engine.connect_to_signal ("EventChanged", self.__event_changed),
            self.__engine.connect_to_signal ("StateChanged", self.__state_changed),
        ]

        self.__engine.GetCurrentState (reply_handler=self.__get_current_state,
                                       error_handler=self._report_dbus_error)
        self.__engine.GetCurrentTrack (reply_handler=self.__get_current_track,
                                       error_handler=self._report_dbus_error)

        self.__poll ()
        self.connected = True
        self.__poll_source = gobject.timeout_add (1000, self.__poll)

    def _disconnect (self):
        for handler in self.__handlers:
            handler.remove ()
        self.__handlers = []

        if self.__poll_source is not None:
            gobject.source_remove (self.__poll_source)
            self.__poll_source = None

        self.__engine = None
        self.__controller = None

        self._set_no_song ()
        self.connected = False

    def toggle_playback (self):
        self.__engine.TogglePlaying (reply_handler=lambda: None,
                                     error_handler=self._report_dbus_error)

    def previous (self):
        self.__controller.Previous (True, reply_handler=lambda: None,
                                        error_handler=self._report_dbus_error)

    def next (self):
        self.__controller.Next (True, reply_handler=lambda: None,
                                    error_handler=self._report_dbus_error)

    # TODO: implement this, once Banshee supports it
    #def rate_song (self, rating):
    #    pass

    def __event_changed (self, event, message, bufferingPercent):
        if event == "startofstream":    #  or event == "trackinfoupdated":  (this is redundant?)
            self.__engine.GetCurrentTrack (reply_handler=self.__get_current_track,
                                           error_handler=self._report_dbus_error)

    def __state_changed (self, state):
        self.playing = (state == "playing")

    def __get_current_state (self, state):
        self.playing = (state == "playing")

    def __get_current_track (self, track):
        self.title = track.get ("name")
        self.artist = track.get ("artist")
        self.album = track.get ("album")
        self.duration = track.get ("length", -1)
        self.rating = track.get ("rating", -1.0)

        # A hack, but Banshee provides no cleaner interface for artwork.
        if self.artist is not None and self.album is not None:
            try:
                art_file = os.path.expanduser ("~/.cache/album-art/%s-%s.jpg" % (self.__cache_name (self.artist),
                                                                                 self.__cache_name (self.album)))
                self.art = gdk.pixbuf_new_from_file (art_file)
            except gobject.GError, message:
                # Happens if file couldn't be loaded
                self.art = None
        else:
            self.art = None

    def __get_position (self, position):
        self.elapsed = position / 1000

    def __poll (self):
        self.__engine.GetPosition (reply_handler=self.__get_position,
                                   error_handler=self._report_dbus_error)
        return True

    def __cache_name (self, name):
        """
        Filter out all non-alphanumeric characters in the name and convert
        it to lowercase.  This is how artists and albums get named in
        Banshee's album art cache directory.
        """

        return "".join ([c for c in name.lower() if c.isalnum()])


def create_instance (conf):
    return Banshee1Plugin (conf)



# Music Applet
# Copyright (C) 2007 Paul Kuliniewicz <paul@kuliniewicz.org>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02111-1301, USA.

"""
Base classes for plugins.

All plugins will ultimately subclass PluginBase, which defines the interface
plugins must adhere to and provides a few convenience methods.  Plugins that
will use D-Bus to communicate with their player should subclass DBusPlugin,
which handles finding when the player of interest is running.

Aside from subclassing one of the base classed defined here, there are a few
other things a plugin module must provide.  Take a look at the example
templates to see what these are.
"""

from gettext import gettext as _
import gobject
import gtk
from gtk import glade

import os.path

import musicapplet.defs
import musicapplet.util


class PluginBase (gobject.GObject):
    """
    Base class for all plugins that implement support for a music player.
    """

    MAX_TIME = 2 ** 31 - 1          # sys.maxint is too big for GObject on 64-bit systems


    ##################################################################
    #
    # Properties
    #
    ##################################################################

    __gproperties__ = {
        "player_name" : (gobject.TYPE_STRING,
                         "player_name",
                         "The (display) name of the player that this plugin connects to",
                         None,
                         gobject.PARAM_READABLE),

        "internal_name" : (gobject.TYPE_STRING,
                           "internal_name",
                           "The internal name of the plugin",
                           None,
                           gobject.PARAM_READABLE),

        "icon_name" : (gobject.TYPE_STRING,
                       "icon_name",
                       "The name of the icon that represents the player",
                       None,
                       gobject.PARAM_READABLE),

        "connected" : (gobject.TYPE_BOOLEAN,
                       "connected",
                       "Whether the plugin is connected",
                       False,
                       gobject.PARAM_READWRITE),

        "playing" : (gobject.TYPE_BOOLEAN,
                     "playing",
                     "Whether the player is currently playing",
                     False,
                     gobject.PARAM_READWRITE),

        "title" : (gobject.TYPE_STRING,
                   "title",
                   "Title of the current song",
                   None,
                   gobject.PARAM_READWRITE),

        "artist" : (gobject.TYPE_STRING,
                    "artist",
                    "Artist who made the current song",
                    None,
                    gobject.PARAM_READWRITE),

        "album" : (gobject.TYPE_STRING,
                   "album",
                   "Album the current song is from",
                   None,
                   gobject.PARAM_READWRITE),

        "duration" : (gobject.TYPE_INT,
                      "duration",
                      "Duration in seconds of the current song",
                      -1, MAX_TIME, -1,
                      gobject.PARAM_READWRITE),

        "rating" : (gobject.TYPE_DOUBLE,
                    "rating",
                    "Rating of the current song",
                    -1.0, 10.0, -1.0,
                    gobject.PARAM_READWRITE),

        "min-rating" : (gobject.TYPE_DOUBLE,
                        "min-rating",
                        "The minimum possible rating",
                        -1.0, 1.0, -1.0,
                        gobject.PARAM_READWRITE),

        "max-rating" : (gobject.TYPE_DOUBLE,
                        "max-rating",
                        "The maximum allowable rating",
                        -1.0, 10.0, -1.0,
                        gobject.PARAM_READWRITE),

        "elapsed" : (gobject.TYPE_INT,
                     "elapsed",
                     "Elapsed time of the current song",
                     -1, MAX_TIME, -1,
                     gobject.PARAM_READWRITE),

        "art": (gobject.TYPE_PYOBJECT,
                "art",
                "Art associated with the song",
                gobject.PARAM_READWRITE),
    }


    player_name = musicapplet.util.make_property ("player_name")
    internal_name = musicapplet.util.make_property ("internal_name")
    icon_name = musicapplet.util.make_property ("icon_name")
    connected = musicapplet.util.make_property ("connected")
    playing = musicapplet.util.make_property ("playing")
    title = musicapplet.util.make_property ("title")
    artist = musicapplet.util.make_property ("artist")
    album = musicapplet.util.make_property ("album")
    duration = musicapplet.util.make_property ("duration")
    rating = musicapplet.util.make_property ("rating")
    min_rating = musicapplet.util.make_property ("min-rating")
    max_rating = musicapplet.util.make_property ("max-rating")
    elapsed = musicapplet.util.make_property ("elapsed")
    art = musicapplet.util.make_property ("art")


    ##################################################################
    #
    # Initialization
    #
    ##################################################################


    def __init__ (self, conf, names, icon_name=None):
        gobject.GObject.__init__ (self)
        self._conf = conf

        if str (type (names)) == "<type 'dict'>":
            player_name = names["display"]
            internal_name = names["internal"]
        else:
            player_name = names
            internal_name = names

        self.__values = {
                "player-name" : player_name,
                "internal-name" : internal_name,
                "icon-name" : icon_name,
                "connected" : False,
                "playing" : False,
                "title" : None,
                "artist" : None,
                "album" : None,
                "duration" : -1,
                "rating" : -1.0,
                "min-rating" : -1.0,
                "max-rating" : -1.0,
                "elapsed" : -1,
                "art": None,
        }


    ##################################################################
    #
    # GObject overrides
    #
    ##################################################################


    def do_get_property (self, property):
        return self.__values[property.name]


    def do_set_property (self, property, value):
        self.__values[property.name] = value


    ##################################################################
    #
    # Public interface
    #
    ##################################################################


    def start (self):
        raise NotImplementedError


    def stop (self):
        raise NotImplementedError


    def toggle_playback (self):
        raise NotImplementedError


    def previous (self):
        raise NotImplementedError


    def next (self):
        raise NotImplementedError


    def launch (self):
        full_key = self._conf.resolve_plugin_key (self, "command")
        command = self._conf.client.get_string (full_key)
        flags = gobject.SPAWN_STDOUT_TO_DEV_NULL | gobject.SPAWN_STDERR_TO_DEV_NULL
        gobject.spawn_async (["/bin/sh", "-c", command], flags=flags)


    # *** OPTIONAL, depending on what the music player or plugin supports ***

    # def rate_song (self, rating):
    #     raise NotImplementedError


    # def create_config_dialog (self):
    #     raise NotImplementedError


    ##################################################################
    #
    # Convenience functions for subclasses
    #
    ##################################################################


    def _browse_for_command (self, parent, program_name=None):
        """
        Run a dialog that prompts for the command to use to launch a program,
        then save it as the plugin's launch command in GConf.
        """

        if program_name is None:
            program_name = self.player_name
        browser = gtk.FileChooserDialog (_("Choose path to %s") % program_name,
                                         parent,
                                         gtk.FILE_CHOOSER_ACTION_OPEN,
                                         (gtk.STOCK_CANCEL, gtk.RESPONSE_CANCEL, gtk.STOCK_OK, gtk.RESPONSE_OK))
        browser.set_default_response (gtk.RESPONSE_OK)
        if browser.run () == gtk.RESPONSE_OK:
            full_key = self._conf.resolve_plugin_key (self, "command")
            self._conf.client.set_string (full_key, browser.get_filename ())
        browser.destroy ()


    def _set_no_song (self):
        """
        Update all the properties to reflect that there is no current song.
        """

        self.playing = False
        self.title = None
        self.artist = None
        self.album = None
        self.duration = -1
        self.elapsed = -1
        self.rating = -1.0
        self.art = None


class DBusPlugin (PluginBase):
    """
    Base class for all plugins that rely on D-Bus to communicate with the
    music player.

    This base class handles establishing all the necessary proxies and
    interfaces that are common for any D-Bus-using plugin.
    """


    def __init__ (self, conf, names, icon_name, dbus_name):
        PluginBase.__init__ (self, conf, names, icon_name)

        import dbus
        import dbus.glib

        if dbus.version < (0, 79, 92):
            raise RuntimeError, _("Requires dbus-python 0.80 or later")

        self.__dbus_name = dbus_name
        self.__bus = None
        self.__bus_handlers = []


    def start (self):
        if self.__bus is not None:
            return

        import dbus
        bus = dbus.SessionBus ()

        proxy = bus.get_object ("org.freedesktop.DBus", "/org/freedesktop/DBus")
        self.__bus = dbus.Interface (proxy, "org.freedesktop.DBus")

        self.__bus_handlers = [
            self.__bus.connect_to_signal ("NameOwnerChanged", self.__name_owner_changed_cb,
                                          arg0=self.__dbus_name)
        ]

        self.__bus.NameHasOwner (self.__dbus_name,
                                 reply_handler=self.__name_has_owner_cb,
                                 error_handler=self._report_dbus_error)


    def stop (self):
        self._disconnect ()
        for handler in self.__bus_handlers:
            handler.remove ()
        self.__bus_handlers = []
        self.__bus = None


    def launch (self):
        if not self.connected:
            full_key = self._conf.resolve_plugin_key (self, "launch")
            if self._conf.client.get_string (full_key) == "Command":
                PluginBase.launch (self)
            else:
                self.__bus.StartServiceByName (self.__dbus_name, 0,
                                               reply_handler=lambda result: None,
                                               error_handler=self._report_dbus_error)


    def create_config_dialog (self):
        xml = self._construct_config_dialog ()
        return xml.get_widget ("dbus-dialog")


    def _construct_config_dialog (self):
        """
        Builds the D-Bus plugin basic configuration dialog, and returns the
        glade.XML object for it.

        You'll want to use this in case you override create_config_dialog to
        add your own custom widgets to the dialog.  The "content" widget is a
        VBox you can add you own widgets to.
        """

        xml = glade.XML (os.path.join (musicapplet.defs.PKG_DATA_DIR, "dbus.glade"))
        dialog = xml.get_widget ("dbus-dialog")

        dialog.set_title (_("%s Plugin") % self.player_name)

        full_key = self._conf.resolve_plugin_key (self, "launch")
        self._conf.bind_string_boolean (full_key, "D-Bus", xml.get_widget ("lu-dbus"), "active")
        self._conf.bind_string_boolean (full_key, "Command", xml.get_widget ("lu-command"), "active")
        self._conf.bind_string_boolean (full_key, "Command", xml.get_widget ("command"), "sensitive")
        self._conf.bind_string_boolean (full_key, "Command", xml.get_widget ("browse"), "sensitive")

        full_key = self._conf.resolve_plugin_key (self, "command")
        self._conf.bind_string (full_key, xml.get_widget ("command"), "text")

        xml.get_widget ("browse").connect ("clicked", lambda button: self._browse_for_command (dialog))
        dialog.set_default_response (gtk.RESPONSE_CLOSE)
        dialog.connect ("response", lambda dialog, response: dialog.hide ())

        return xml


    def __name_has_owner_cb (self, has_owner):
        if has_owner:
            self._connect ()


    def __name_owner_changed_cb (self, name, old_owner, new_owner):
        if new_owner != "":
            self._connect ()
        else:
            self._disconnect ()


    def _report_dbus_error (self, error):
        """
        Callback for reporting D-Bus errors.
        """
        
        # TODO: Should this be visible to the user?
        print "D-Bus error: %s" % error


    def _connect (self):
        """
        Called when the player has appeared on the bus.
        """

        raise NotImplementedError


    def _disconnect (self):
        """
        Called when the player has disappeared from the bus.
        """

        raise NotImplementedError



class MPRISPlugin (DBusPlugin):
    """
    Base class for all plugins that communicate with the music player
    using the MPRIS interface via D-Bus.
    """

    CAN_GO_NEXT          = 1 << 0
    CAN_GO_PREV          = 1 << 1
    CAN_PAUSE            = 1 << 2
    CAN_PLAY             = 1 << 3
    CAN_SEEK             = 1 << 4
    CAN_PROVIDE_METADATA = 1 << 5

    STATUS_PLAYING = 0
    STATUS_PAUSED  = 1
    STATUS_STOPPED = 2


    def __init__ (self, conf, names, icon_name, dbus_name):
        DBusPlugin.__init__ (self, conf, names, icon_name, dbus_name)

        self.__dbus_name = dbus_name
        self.__player = None
        self.__handlers = []
        self.__status = self.STATUS_STOPPED
        self.__caps = 0
        self.__poll_source = None


    def _connect (self):
        if self.__player is not None:
            return

        import dbus
        bus = dbus.SessionBus ()

        proxy = bus.get_object (self.__dbus_name, "/Player")
        self.__player = dbus.Interface (proxy, "org.freedesktop.MediaPlayer")

        self.__handlers = [
            self.__player.connect_to_signal ("TrackChange", self.__get_metadata),
            self.__player.connect_to_signal ("StatusChange", self.__get_status),
            self.__player.connect_to_signal ("CapsChange", self.__get_caps),
        ]

        self.__player.GetMetadata (reply_handler=self.__get_metadata,
                                   error_handler=self._report_dbus_error)

        self.__player.GetStatus (reply_handler=self.__get_status,
                                 error_handler=self._report_dbus_error)

        self.__player.GetCaps (reply_handler=self.__get_caps,
                               error_handler=self._report_dbus_error)

        self.__player.PositionGet (reply_handler=self.__position_get,
                                   error_handler=self._report_dbus_error)

        self.__poll_source = gobject.timeout_add (1000, self.__poll)

        self.connected = True


    def _disconnect (self):
        for handler in self.__handlers:
            handler.remove ()
        self.__handlers = []

        if self.__poll_source is not None:
            gobject.source_remove (self.__poll_source)
            self.__poll_source = None

        self.__player = None

        self._set_no_song ()
        self.connected = False


    def toggle_playback (self):
        if self.__status == self.STATUS_STOPPED:
            self.__player.Play ()
        else:
            self.__player.Pause ()


    def previous (self):
        self.__player.Prev ()


    def next (self):
        self.__player.Next ()


    def __get_metadata (self, metadata):
        print "DEBUG: MPRIS %s: __get_metadata:" % self.__dbus_name
        for key in metadata:
            print "DEBUG:    ", key, "=>", metadata[key]

        self.title = metadata.get ("title", None) or None
        self.artist = metadata.get ("artist", None) or None
        self.album = metadata.get ("album", None) or None

        if metadata.has_key ("time"):
            self.duration = metadata.get ("time")
        elif metadata.has_key ("mtime"):
            self.duration = metadata.get ("mtime") / 1000
        elif metadata.has_key ("length"):
            self.duration = metadata.get ("length") / 1000
        else:
            self.duration = -1

        self.rating = metadata.get ("rating", -1.0)

        if metadata.has_key ("arturl"):
            path = metadata["arturl"]
            if path.startswith ("file://"):
                path = path[7:]
            self.art = gdk.pixbuf_new_from_file (path)
        else:
            self.art = None


    def __get_status (self, status):
        print "DEBUG: MPRIS %s: __get_status:" % self.__dbus_name, status

        # XXX: The spec (http://wiki.xmms2.xmms.se/index.php/MPRIS) claims
        #      this is a list of four ints, but some players only bother
        #      with the first one

        import dbus
        if type (status) == dbus.Struct:
            self.__status = status[0]
        else:
            self.__status = status
        self.playing = (self.__status == self.STATUS_PLAYING)
        if self.__status == self.STATUS_STOPPED:
            self._set_no_song ()


    def __get_caps (self, caps):
        print "DEBUG: MPRIS %s: __get_caps:" % self.__dbus_name, caps
        self.__caps = caps
        if not (caps & self.CAN_PROVIDE_METADATA):
            self._set_no_song ()


    def __position_get (self, elapsed):
        print "DEBUG: MPRIS %s: __position_get:" % self.__dbus_name, elapsed
        if elapsed > 0 or self.__status != self.STATUS_STOPPED:
            self.elapsed = elapsed / 1000


    def __poll (self):
        self.__player.PositionGet (reply_handler=self.__position_get,
                                   error_handler=self._report_dbus_error)
        return True

    # def rate_song (self, rating): ...


gobject.type_register (PluginBase)
gobject.type_register (DBusPlugin)
gobject.type_register (MPRISPlugin)



# Music Applet
# Copyright (C) 2007 Paul Kuliniewicz <paul@kuliniewicz.org>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02111-1301, USA.

from gettext import gettext as _
import gobject
import gtk
from gtk import gdk, glade

import os

import musicapplet.defs
import musicapplet.player


INFO = {
        "name": "XMMS2",
        "version": musicapplet.defs.VERSION,
        "icon-name": "music-applet-xmms2",
        "author": "Paul Kuliniewicz <paul@kuliniewicz.org>",
        "copyright": "(C) 2007 Paul Kuliniewicz",
        "website": "http://www.kuliniewicz.org/music-applet/",
}


class XMMS2Plugin (musicapplet.player.PluginBase):
    """
    Music Applet plugin to interface with XMMS2.
    """


    ##################################################################
    #
    # Creation and overall management
    #
    ##################################################################


    def __init__ (self, conf):
        musicapplet.player.PluginBase.__init__ (self, conf, INFO["name"], INFO["icon-name"])

        import xmmsclient

        self.min_rating = 0.0
        self.max_rating = 5.0

        self.__client = xmmsclient.XMMS ("MusicApplet")
        self.__connector = None
        self.__poll_source = None

        self.__current_id = 0
        self.__art_url = None


    def start (self):
        if not self.connected and self.__poll_source is None and self.__poll ():
            self.__poll_source = gobject.timeout_add (1000, self.__poll)


    def stop (self):
        if self.connected:
            self.__disconnect (False)

        if self.__poll_source is not None:
            gobject.source_remove (self.__poll_source)
            self.__poll_source = None


    def create_config_dialog (self):
        xml = glade.XML (os.path.join (musicapplet.defs.PKG_DATA_DIR, "xmms2.glade"))
        dialog = xml.get_widget ("xmms2-dialog")

        full_key = self._conf.resolve_plugin_key (self, "conn_path")
        self._conf.bind_string (full_key, xml.get_widget ("path"), "text")

        full_key = self._conf.resolve_plugin_key (self, "command")
        self._conf.bind_string (full_key, xml.get_widget ("command"), "text")

        xml.get_widget ("browse").connect ("clicked",
                    lambda button: self._browse_for_command (dialog, _("%s Launcher") % "XMMS2"))
        dialog.set_title (_("%s Plugin") % "XMMS2")
        dialog.set_default_response (gtk.RESPONSE_CLOSE)
        dialog.connect ("response", lambda dialog, response: dialog.hide ())
        return dialog


    def __poll (self):
        try:
            full_key = self._conf.resolve_plugin_key (self, "conn_path")
            path = self._conf.client.get_string (full_key)
            if path is None:
                path = os.getenv ("XMMS_PATH")
            self.__client.connect (path, self.__disconnect)
            self.__connector = GLibConnector (self.__client)

            self.__client.broadcast_playback_current_id (self.__set_current_id)
            self.__client.broadcast_medialib_entry_changed (self.__entry_changed)
            self.__client.signal_playback_playtime (
                    lambda result: (self.__set_elapsed (result), result.restart ()))
            self.__client.broadcast_playback_status (self.__set_status)

            self.__client.playback_current_id (self.__set_current_id)
            self.__client.playback_playtime (self.__set_elapsed)
            self.__client.playback_status (self.__set_status)

            self.connected = True
            self.__poll_source = None
            return False

        except IOError:
            return True


    def __disconnect (self, still_enabled=True):
        self.__connector.shutdown ()
        self.__connector = None
        self._set_no_song ()
        self.connected = False

        if still_enabled and self.__poll_source is None:
            self.__poll_source = gobject.timeout_add (1000, self.__poll)


    ##################################################################
    #
    # Playback control
    #
    ##################################################################


    def toggle_playback (self):
        if self.playing:
            self.__client.playback_pause ()
        else:
            self.__client.playback_start ()


    def previous (self):
        self.__client.playlist_set_next_rel (-1, lambda result: self.__client.playback_tickle ())


    def next (self):
        self.__client.playlist_set_next_rel (1, lambda result: self.__client.playback_tickle ())


    def rate_song (self, rating):
        self.__client.medialib_property_set (self.__current_id, "rating", int (rating), "clients/generic")
        self.rating = rating


    ##################################################################
    #
    # Status information
    #
    ##################################################################


    def __set_current_id (self, result):
        self.__current_id = result.value ()
        self.__client.medialib_get_info (self.__current_id, self.__set_song_info)


    def __entry_changed (self, result):
        if result.value () == self.__current_id:
            self.__client.medialib_get_info (self.__current_id, self.__set_song_info)


    def __set_elapsed (self, result):
        self.elapsed = result.value () / 1000


    def __set_status (self, result):
        import xmmsclient
        self.playing = (result.value () == xmmsclient.PLAYBACK_STATUS_PLAY)


    def __set_song_info (self, result):
        info = result.value ()
        if info is None:
            self._set_no_song ()
            return

        if "title" in info:
            if self.title != info["title"]:
                self.title = info["title"]
        else:
            self.title = None

        if "artist" in info:
            if self.artist != info["artist"]:
                self.artist = info["artist"]
        else:
            self.artist = None

        if "album" in info:
            if self.album != info["album"]:
                self.album = info["album"]
        else:
            self.album = None

        if "duration" in info:
            self.duration = info["duration"] / 1000
        else:
            self.duration = -1

        vote_score = ("clients/generic/voting", "vote_score")
        vote_count = ("clients/generic/voting", "vote_count")
        if ("clients/generic", "rating") in info:
            self.rating = info[("clients/generic", "rating")]
        elif vote_score in info and vote_count in info and info[vote_count] > 0:
            self.rating = info[vote_score] / info[vote_count]
        else:
            self.rating = 0.0

        art_url = self.__find_art_url (info)
        if self.__art_url != art_url:
            self.__art_url = art_url
            if art_url is not None:
                if art_url.startswith ("file://"):
                    self.art = gdk.pixbuf_new_from_file (art_url[7:])
                # TODO: A way to pull it from http:// or whatever else might be used
            else:
                self.art = None


    def __find_art_url (self, info):
        for source in ["clients/generic/arturl", "clients/generic/arturl_unverified"]:
            for key in ["album_front_thumbnail", "album_front_small", "album_front_large"]:
                if (source, key) in info:
                    return info[(source, key)]
        return None


def create_instance (conf):
    return XMMS2Plugin (conf)


##################################################################
#
# The following was taken from the xmms2-tutorial git repo
#
# http://git.xmms.se/?p=xmms2-tutorial.git;a=summary
#
##################################################################

####
#
# A GLib connector for PyGTK - 
#	to use with your cool PyGTK xmms2 client.
# Tobias Rundstrom <tru@xmms.org)
#       ... and modified by Paul Kuliniewicz <paul@kuliniewicz.org>
#
# Just create the GLibConnector() class with a xmmsclient as argument
#
###

class GLibConnector:
    def __init__ (self, xmms):
        self.xmms = xmms
        xmms.set_need_out_fun (self.need_out)
        self.in_source = gobject.io_add_watch (self.xmms.get_fd (), gobject.IO_IN, self.handle_in)
        self.out_source = None

    def need_out (self, i):
        if self.xmms.want_ioout () and self.out_source is None:
            self.out_source = gobject.io_add_watch (self.xmms.get_fd (), gobject.IO_OUT, self.handle_out)

    def handle_in (self, source, cond):
        if cond == gobject.IO_IN:
            self.xmms.ioin ()

        return True

    def handle_out (self, source, cond):
        if cond == gobject.IO_OUT:
            self.xmms.ioout ()
        if not self.xmms.want_ioout ():
            self.out_source = None
            return False
        else:
            return True

    def shutdown (self):
        if self.in_source is not None:
            gobject.source_remove (self.in_source)
            self.in_source = None
        if self.out_source is not None:
            gobject.source_remove (self.out_source)
            self.out_source = None
        self.xmms.set_need_out_fun (None)



# Music Applet
# Copyright (C) 2008 Paul Kuliniewicz <paul@kuliniewicz.org>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02111-1301, USA.

import musicapplet.defs
import musicapplet.player


INFO = {
        "name": "VLC",
        "version": musicapplet.defs.VERSION,
        "icon-name": "vlc",
        "author": "Paul Kuliniewicz <paul@kuliniewicz.org>",
        "copyright": "(C) 2008 Paul Kuliniewicz",
        "website": "http://www.kuliniewicz.org/music-applet/",
}


class VLCPlugin (musicapplet.player.MPRISPlugin):
    """
    Music Applet plugin to interface with VLC.
    """

    def __init__ (self, conf):
        musicapplet.player.MPRISPlugin.__init__ (self, conf,
                                                 INFO["name"], INFO["icon-name"],
                                                 "org.mpris.vlc")


def create_instance (conf):
    return VLCPlugin (conf)



# Music Applet
# Copyright (C) 2007 Paul Kuliniewicz <paul@kuliniewicz.org>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02111-1301, USA.

"""
Classes that manage the collection of available plugins.

Plugins shouldn't need to worry about anything here.
"""

from gettext import gettext as _
import gconf
import gobject
import gtk
from gtk import glade

import os
import pydoc
from xml.sax import saxutils

import musicapplet.defs
import musicapplet.util


class Manager (gtk.ListStore):
    """
    Manages the collection of available plugins.
    """

    COL_PATH = 0
    COL_INTERNAL_NAME = 1
    COL_ICON_NAME = 2
    COL_MODULE = 3
    COL_OBJECT = 4
    COL_ENABLED = 5
    COL_HANDLER = 6
    COL_ERROR = 7
    COL_DIALOG = 8
    COL_DISPLAY_NAME = 9


    __gproperties__ = {
        "active": (gobject.TYPE_PYOBJECT,
                   "active",
                   "The currently active plugin instance",
                   gobject.PARAM_READWRITE),

        "preferred": (gobject.TYPE_PYOBJECT,
                      "preferred",
                      "The preferred plugin instance",
                      gobject.PARAM_READWRITE),
    }

    active = musicapplet.util.make_property ("active")
    preferred = musicapplet.util.make_property ("preferred")


    def __init__ (self, conf, dirs):
        gtk.ListStore.__init__ (self, str,                      # COL_PATH
                                      str,                      # COL_INTERNAL_NAME
                                      str,                      # COL_ICON_NAME
                                      gobject.TYPE_PYOBJECT,    # COL_MODULE
                                      gobject.TYPE_PYOBJECT,    # COL_OBJECT
                                      bool,                     # COL_ENABLED
                                      int,                      # COL_HANDLER
                                      str,                      # COL_ERROR
                                      gobject.TYPE_PYOBJECT,    # COL_DIALOG
                                      str)                      # COL_DISPLAY_NAME
        self.__active = None
        self.__preferred = None
        self.__conf = conf

        self.__dirs = dirs
        avail = {}
        for dir in dirs:
            try:
                for file in os.listdir (dir):
                    try:
                        path = os.path.join (dir, file)
                        module = pydoc.importfile (path)
                        module.INFO["path"] = path
                        name = module.INFO.get ("internal-name") or module.INFO["name"]
                        if name not in avail.keys () or self.__better_plugin (module, avail[name]):
                            avail[name] = module
                    except Exception, detail:
                        if not file.startswith ("__init__.py"):
                            print "Failed to open module at %s: %s" % (path, detail)
            except OSError, detail:
                if dir.find ("/.gnome2/music-applet/plugins") == -1:
                    print "Failed to scan %s: %s" % (dir, detail)

        names = avail.keys ()
        names.sort (key=str.lower)
        for name in names:
            module = avail[name]
            self.append ((module.INFO["path"],
                          module.INFO.get ("internal-name") or module.INFO["name"],
                          module.INFO["icon-name"],
                          module,
                          None,
                          False,
                          0,
                          "",
                          None,
                          module.INFO.get ("display-name") or module.INFO["name"]))

        full_key = conf.resolve_key ("enabled_plugins")
        self.foreach (self.__set_enabled, (conf.client.get_list (full_key, gconf.VALUE_STRING), False))
        conf.client.notify_add (full_key,
            lambda client, id, entry, unused:
                   entry.value is not None and
                           self.foreach (self.__set_enabled,
                                         ([v.get_string () for v in entry.value.get_list ()], False)))

        full_key = conf.resolve_key ("preferred_player")
        self.__update_preferred (conf.client.get_string (full_key))
        conf.client.notify_add (full_key,
            lambda client, id, entry, unused:
                   entry.value is not None and self.__update_preferred (entry.value.get_string ()))

        self.foreach (self.__update_status, None)
        self.foreach (self.__find_connected)


    def do_get_property (self, property):
        if property.name == "active":
            return self.__active
        elif property.name == "preferred":
            return self.__preferred
        else:
            raise AttributeError, "unknown property '%s'" % property.name


    def do_set_property (self, property, value):
        if property.name == "active":
            self.foreach (self.__update_status, value)
            self.__active = value
        elif property.name == "preferred":
            self.__preferred = value
        else:
            raise AttributeError, "unknown property '%s'" % property.name


    def toggle_plugin (self, path):
        """
        Toggle the enabled status of the plugin specified by the given path.
        """

        iter = self.get_iter (path)
        if self.get_value (iter, self.COL_ENABLED):
            self.__disable_plugin (iter)
        else:
            self.__enable_plugin (iter, True)


    def __enable_plugin (self, iter, activate):
        """
        Attempt to enable a plugin.  If the operation fails, the plugin will
        remain unenabled.
        """

        (internal_name, module, object, enabled) = self.get (
                iter, self.COL_INTERNAL_NAME, self.COL_MODULE, self.COL_OBJECT, self.COL_ENABLED)
        if enabled:
            return

        try:
            if object is None:
                object = module.create_instance (self.__conf)
                handler = object.connect ("notify::connected", self.__plugin_connected_cb)
                self.set (iter, self.COL_OBJECT, object,
                                self.COL_HANDLER, handler,
                                self.COL_ICON_NAME, module.INFO["icon-name"],
                                self.COL_ERROR, "")
        
            self.set_value (iter, self.COL_ENABLED, True)
            if self.__active is None:
                object.start ()
                if activate and object.connected:
                    self.active = object

            full_key = self.__conf.resolve_key ("enabled_plugins")
            enabled_names = self.__conf.client.get_list (full_key, gconf.VALUE_STRING)
            if internal_name not in enabled_names:
                enabled_names.append (internal_name)
                self.__conf.client.set_list (full_key, gconf.VALUE_STRING, enabled_names)

        except Exception, error:
            # XXX: Can we rely on the underlying icon name of the stock icon?
            self.set (iter, self.COL_ICON_NAME, "stock_dialog-error", self.COL_ERROR, error)


    def __disable_plugin (self, iter):
        """
        Disable a plugin.
        """

        (internal_name, object, enabled) = self.get (iter, self.COL_INTERNAL_NAME, self.COL_OBJECT, self.COL_ENABLED)
        if not enabled:
            return

        self.set (iter, self.COL_ENABLED, False)
        if object is not None:
            object.stop ()
            if self.__active is object:
                self.active = None
                self.foreach (self.__find_connected)
            if self.__preferred is object:
                self.preferred = None
                full_key = self.__conf.resolve_key ("preferred_player")
                self.__conf.client.set_string (full_key, "")

        full_key = self.__conf.resolve_key ("enabled_plugins")
        enabled_names = self.__conf.client.get_list (full_key, gconf.VALUE_STRING)
        if internal_name in enabled_names:
            enabled_names.remove (internal_name)
            self.__conf.client.set_list (full_key, gconf.VALUE_STRING, enabled_names)


    def __set_enabled (self, model, path, iter, data):
        """
        TreeModel foreach callback that changes plugins' enabled status to
        match the given list of plugin names.
        """

        (enabled_names, activate) = data
        internal_name = model.get_value (iter, model.COL_INTERNAL_NAME)
        if internal_name in enabled_names:
            self.__enable_plugin (iter, activate)
        else:
            self.__disable_plugin (iter)

        return False


    def __better_plugin (self, new, old):
        """
        Check if one module is better than another one.

        Plugins are compared by the following rules, in order:
          * Later versions are better than older versions
          * More optimized copies are better than less optimized copies
          * The new one wins
        """

        new_ver = new.INFO["version"].split (".")
        old_ver = old.INFO["version"].split (".")
        if new_ver > old_ver:
            return True
        if new_ver < old_ver:
            return False

        new_opt = self.__optimization (new)
        old_opt = self.__optimization (old)
        if new_opt > old_opt:
            return True
        if new_opt < old_opt:
            return False

        return True


    def __optimization (self, module):
        """
        Get an estimate of the optimization level of a module.

        The higher the result, the more optimized the module is; unknown
        extensions are assumed to be native code.
        """

        levels = {"pyo": 3, "pyc": 2, "py": 1}
        name = module.INFO["path"]
        ext = name[name.rindex (".") + 1 : -1]
        if ext in levels.keys ():
            return levels[ext]
        else:
            return 4


    def __update_status (self, model, path, iter, active):
        """
        A gtk.TreeModel foreach callback for updating rows when the active
        plugin changes.
        """

        (object, enabled) = model.get (iter, model.COL_OBJECT, model.COL_ENABLED)
        if active is not None:
            if object is not None and object is not active:
                object.stop ()
        else:
            if enabled and object is not None:
                object.start ()

        return False


    def __find_connected (self, model, path, iter):
        """
        A gtk.TreeModel foreach callback that looks for a row with a connected
        plugin.
        """

        (object, enabled) = model.get (iter, model.COL_OBJECT, model.COL_ENABLED)
        if enabled and object is not None and object.connected:
            self.__plugin_connected_cb (object, None)
            return True
        else:
            return False


    def __update_preferred (self, internal_name):
        """
        Update the preferred property based on the name of the preferred
        player.
        """

        iter = musicapplet.util.search_model (self, self.COL_INTERNAL_NAME, internal_name)
        if iter is not None:
            self.preferred = self.get_value (iter, self.COL_OBJECT)


    def __plugin_connected_cb (self, object, pspec):
        if object.connected and self.__active is None:
            self.active = object
        elif not object.connected and object is self.__active:
            self.active = None


def create_plugins_dialog (manager, conf):
    """
    Create a dialog that lets the user configure plugins.
    """

    def selection_changed_cb (selection):
        keys = ["version", "author", "copyright", "website"]
        (model, iter) = selection.get_selected ()
        if iter is not None:
            (module, object, enabled, error) = model.get (
                    iter, model.COL_MODULE, model.COL_OBJECT, model.COL_ENABLED, model.COL_ERROR)
            xml.get_widget ("name").set_markup ("<big><b>" +
                                                _("%s Plugin") % saxutils.escape (module.INFO.get ("display-name") or module.INFO["name"]) +
                                                "</b></big>")
            for key in keys:
                xml.get_widget (key).set_text (module.INFO[key])
            xml.get_widget ("configure").set_sensitive (enabled and object is not None and "create_config_dialog" in dir (object))
            if error != "":
                xml.get_widget ("error").set_text (error)
                xml.get_widget ("error-box").show ()
            else:
                xml.get_widget ("error").set_text ("")
                xml.get_widget ("error-box").hide ()
        else:
            xml.get_widget ("name").set_markup ("<big><b> </b></big>")
            for key in keys:
                xml.get_widget (key).set_text ("")
            xml.get_widget ("configure").set_sensitive (False)
            xml.get_widget ("error").set_text ("")
            xml.get_widget ("error-box").hide ()

    def row_changed_cb (model, path, iter):
        selection = xml.get_widget ("plugins").get_selection ()
        if selection.iter_is_selected (iter):
            selection_changed_cb (selection)

    def configure_cb (button):
        selection = xml.get_widget ("plugins").get_selection ()
        (model, iter) = selection.get_selected ()
        if iter is not None:
            (object, dialog) = model.get (iter, model.COL_OBJECT, model.COL_DIALOG)
            if dialog is None:
                dialog = object.create_config_dialog ()
                model.set_value (iter, model.COL_DIALOG, dialog)
                dialog.connect ("destroy", lambda dialog: model.set_value (iter, model.COL_DIALOG, None))
                dialog.show ()
            else:
                dialog.present ()

    xml = glade.XML (os.path.join (musicapplet.defs.PKG_DATA_DIR, "plugins.glade"))

    # Plugins list

    plugins = xml.get_widget ("plugins")
    plugins.set_model (manager)
    manager.connect ("row-changed", row_changed_cb)

    renderer = gtk.CellRendererToggle ()
    renderer.set_property ("activatable", True)
    renderer.connect ("toggled", lambda renderer, path: manager.toggle_plugin (path))
    column = gtk.TreeViewColumn ("Enabled", renderer, active=manager.COL_ENABLED)
    plugins.append_column (column)

    renderer = gtk.CellRendererPixbuf ()
    column = gtk.TreeViewColumn ("Icon", renderer, icon_name=manager.COL_ICON_NAME)
    plugins.append_column (column)

    renderer = gtk.CellRendererText ()
    column = gtk.TreeViewColumn ("Name", renderer, text=manager.COL_DISPLAY_NAME)
    plugins.append_column (column)

    selection = plugins.get_selection ()
    selection.connect ("changed", selection_changed_cb)
    selection.select_path ("0")

    # Preferred plugin combo box

    preferred = xml.get_widget ("preferred")
    model = manager.filter_new ()
    model.set_visible_column (manager.COL_ENABLED)
    preferred.set_model (model)

    renderer = gtk.CellRendererPixbuf ()
    preferred.pack_start (renderer, False)
    preferred.set_attributes (renderer, icon_name=manager.COL_ICON_NAME)

    renderer = gtk.CellRendererText ()
    preferred.pack_start (renderer, True)
    preferred.set_attributes (renderer, text=manager.COL_DISPLAY_NAME)

    full_key = conf.resolve_key ("preferred_player")
    conf.bind_combo_string (full_key, preferred, manager.COL_INTERNAL_NAME)

    # Everything else

    xml.get_widget ("configure").connect ("clicked", configure_cb)

    xml.get_widget ("plugins-dialog").set_default_response (gtk.RESPONSE_CLOSE)
    xml.get_widget ("plugins-dialog").connect ("response", lambda dialog, response: dialog.hide ())
    return xml.get_widget ("plugins-dialog")


gobject.type_register (Manager)
